# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the request handlers of an-website."""

from __future__ import annotations

import itertools
import urllib.parse
from http.cookies import SimpleCookie
from urllib.parse import urlencode

import orjson as json
from lxml.html.html5parser import HTMLParser

from . import (  # noqa: F401  # pylint: disable=unused-import
    FetchCallable,
    app,
    assert_url_query,
    assert_valid_response,
    fetch,
)


async def test_setting_stuff_without_cookies(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Test changing settings with requests without saving to cookie."""
    for request_body in (
        json.dumps(
            {
                "bumpscosity": "0",
                "theme": "pink",
                "no_3rd_party": "sure",
                "dynload": "sure",
                "openmoji": "img",
            }
        ),
        json.dumps(
            {
                "bumpscosity": "0",
                "theme": "pink",
                "no_3rd_party": "sure",
                "dynload": "sure",
                "openmoji": "img",
                "save_in_cookie": "nope",
            }
        ),
        "theme=pink&no_3rd_party=sure&dynload=sure&openmoji=img&bumpscosity=0",
        (
            "theme=pink&no_3rd_party=s&dynload=s&save_in_cookie=n&openmoji=i"
            "&bumpscosity=0"
        ),
        "",
    ):
        response = await fetch(
            (
                "/einstellungen"
                if request_body
                else (
                    "/einstellungen?theme=pink&no_3rd_party=sure&dynload=sure"
                    "&openmoji=img&bumpscosity=0&save_in_cookie=nope"
                )
            ),
            method="POST" if request_body else "GET",
            headers=(
                {"Content-Type": "application/json"}
                if isinstance(request_body, bytes)
                else None
            ),
            body=request_body or None,
            follow_redirects=True,
        )
        assert_valid_response(response, content_type="text/html;charset=utf-8")

        assert_url_query(
            response.effective_url,
            theme="pink",
            no_3rd_party="sure",
            dynload="sure",
            openmoji="img",
            save_in_cookie="nope",
            bumpscosity="0",
        )

        root = HTMLParser(namespaceHTMLElements=False).parse(response.body)
        home_url = root.xpath("//*[@id='back-to-home']")[0].get("href")

        assert_url_query(
            home_url,
            theme="pink",
            no_3rd_party="sure",
            dynload="sure",
            openmoji="img",
            save_in_cookie=None,
            bumpscosity="0",
        )


def parse_cookie(cookie: str) -> SimpleCookie:
    """Parse a cookie string to a SimpleCookie."""
    simple_cookie = SimpleCookie()
    simple_cookie.load(cookie)
    return simple_cookie


async def test_setting_stuff_and_saving_to_cookies(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Test changing settings with requests with saving to cookie."""
    for request_body in (
        json.dumps(
            {
                "theme": "pink",
                "no_3rd_party": "sure",
                "dynload": "sure",
                "save_in_cookie": "sure",
                "openmoji": "img",
                "bumpscosity": "0",
            }
        ),
        "theme=pink&no_3rd_party=s&dynload=s&save_in_cookie=s&openmoji=i&bumpscosity=0",
    ):
        response = await fetch(
            "/einstellungen",
            method="POST",
            headers=(
                {"Content-Type": "application/json"}
                if isinstance(request_body, bytes)
                else None
            ),
            body=request_body,
        )
        if request_body:
            assert_valid_response(
                response, content_type="text/html;charset=utf-8"
            )
        else:
            assert response.headers["Location"].endswith("/einstellungen")

        for cookie_header in response.headers.get_list("Set-Cookie"):
            cookie = parse_cookie(cookie_header)
            assert len(cookie) == 1
            morsel = cookie[tuple(cookie)[0]]
            assert morsel["expires"]
            assert morsel["path"] == "/"
            assert morsel["samesite"] == "Strict"
            assert morsel.key in {
                "advanced_settings",
                "ask_before_leaving",
                "bumpscosity",
                "compat",
                "dynload",
                "effects",
                "no_3rd_party",
                "openmoji",
                "theme",
                "scheme",
            }
            if morsel.key == "theme":
                assert morsel.value == "pink"
            elif morsel.key == "scheme":
                assert morsel.value == "system"
            elif morsel.key == "openmoji":
                assert morsel.value == "img"
            elif morsel.key == "bumpscosity":
                assert morsel.value == "0"
            elif morsel.key in {
                "advanced_settings",
                "compat",
                "ask_before_leaving",
            }:
                assert morsel.value == "nope"
            else:
                assert morsel.value == "sure"


async def test_setting_stuff_and_saving_to_cookies2(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Test changing settings with requests with saving to cookie."""
    options: tuple[dict[str, str | None], ...] = (
        {
            "theme": "christmas",
            "no_3rd_party": "nope",
            "dynload": "nope",
            "openmoji": "glyf_colr1",
            "bumpscosity": "76",
            "advanced_settings": "nope",
            "compat": "nope",
            "access_token": "xyzzy",
            "ask_before_leaving": "nope",
            "effects": "nope",
            "scheme": "dark",
        },
        {
            "theme": "pink",
            "no_3rd_party": "sure",
            "dynload": "sure",
            "openmoji": "img",
            "bumpscosity": "0",
            "advanced_settings": None,
            "compat": "sure",
            "ask_before_leaving": "sure",
            "effects": "sure",
            "scheme": "system",
        },
        {
            "bumpscosity": "100",
            "theme": "default",
            "no_3rd_party": "nope",
            "dynload": None,
            "openmoji": "img",
            "advanced_settings": "sure",
            "compat": "nope",
            "ask_before_leaving": "nope",
            "effects": "nope",
            "scheme": "light",
        },
    )
    cookies: list[str | None] = [
        "; ".join([f"{key}={value}" for key, value in data.items()])
        for data in options
    ]
    cookies.append(None)

    for data, req_cookie in itertools.product(options, cookies):
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        if req_cookie is not None:
            headers["Cookie"] = req_cookie
        response = await fetch(
            "/einstellungen",
            method="POST",
            headers=headers,
            body=urlencode(
                {
                    "save_in_cookie": "sure",
                    **{
                        key: value
                        for key, value in data.items()
                        if value is not None
                    },
                }
            ),
        )
        assert_valid_response(response, content_type="text/html;charset=utf-8")

        root = HTMLParser(namespaceHTMLElements=False).parse(response.body)
        home_url = root.xpath("//*[@id='back-to-home']")[0].get("href")

        if query := urllib.parse.urlsplit(home_url).query:
            raise AssertionError(f"Back to home URL has query: {query}")

        for cookie_header in response.headers.get_list("Set-Cookie"):
            cookie = parse_cookie(cookie_header)
            assert len(cookie) == 1
            morsel = cookie[tuple(cookie)[0]]
            assert morsel["expires"]
            assert morsel["path"] == "/"
            assert morsel["samesite"] == "Strict"
            assert (
                morsel.key in data
            ), f"{morsel.key} not in {list(data.keys())}"
            if morsel.key == "access_token":
                assert morsel.value == "eHl6enk="
            else:
                assert morsel.value == (
                    data[morsel.key] or "nope"
                ), f"{morsel.value} != {data[morsel.key]}"



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A page with a list of services that are cool and hosted by us."""

from __future__ import annotations

import dataclasses

from ..utils.request_handler import HTMLRequestHandler
from ..utils.utils import ModuleInfo


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(("/services", ServicesHandler),),
        name="Service-Liste",
        description="Liste von coolen Services des Asozialen Netzwerks",
        path="/services",
        keywords=("Service", "Liste"),
        aliases=("/services-list",),
    )


@dataclasses.dataclass(slots=True)
class Service:
    """A class representing a service."""

    title: str
    text: str
    infos: None | dict[str, str]


SERVICES: tuple[Service, ...] = (
    Service(
        "Minceraft-Server",
        (
            "Der Survival-Minceraft-Server des Asozialen Netzwerks funktioniert"
            " auch ohne einen Minceraft-Account und hat dafür eine Integration"
            " mit ely.by, damit auch Spieler, die ihren Minceraft-Account nicht"
            " verraten möchten, einen eigenen Skin nutzen können."
        ),
        {
            "Domain": "minceraft.asozial.org",
            "Version": "1.15.2 (unterstützt 1.7-1.20)",
            "Karte": "https://minceraft.asozial.org",
        },
    ),
    Service(
        "SuperTuxKart-Server",
        (
            "Der SuperTuxKart-Server des Asozialen Netzwerks ist durchgehend "
            "online."
        ),
        {
            "Name": "Das Asoziale Netzwerk",
            "SuperTuxKart-Download": "https://supertuxkart.net/Download",
        },
    ),
    Service(
        "Syncplay-Server",
        (
            "Mit dem Syncplay-Server des Asozialen Netzwerks kann man Online "
            "zusammen Sachen gucken."
        ),
        {
            "Domain": "syncplay.asozial.org:8999",
            "Installations-Guide": "https://syncplay.pl/guide/install",
        },
    ),
    # Service(
    #     "Matrix-Heimserver",
    #     (
    #         "Der Matrix-Heimserver des Asozialen Netzwerks ist zuverlässig und"
    #         " nutzt im Gegensatz zu den meisten anderen Servern Dendrite"
    #         " anstatt Synapse. Die Erstellung eines Accounts ist 100% kostenlos"
    #         " und ohne E-Mail-Adresse oder Telefonnummer möglich."
    #     ),
    #     {
    #         "Domain": "asozial.org",
    #         "Matrix-Client": "https://chat.asozial.org",
    #     },
    # ),
    Service(
        "XMPP-Server",
        "Das Asoziale Netzwerk hat einen XMPP-Server.",
        {
            "Domain": "asozial.org",
            "Software": "ejabberd",
        },
    ),
)


class ServicesHandler(HTMLRequestHandler):
    """The request handler for this page."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the service list page."""
        if head:
            return
        await self.render(
            "pages/services.html",
            services=SERVICES,
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The random text API of the website."""

from __future__ import annotations

from dataclasses import dataclass
from random import Random
from typing import ClassVar

import regex

from ..utils.data_parsing import parse_args
from ..utils.request_handler import APIRequestHandler
from ..utils.utils import ModuleInfo
from . import DIR


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=((r"/api/zufaelliger-text", RandomText),),
        name="Zufälliger Text",
        description="Generiere zu jedem Request einen zufälligen Text.",
        path="/api/zufaelliger-text",
        hidden=True,
    )


@dataclass(slots=True)
class Arguments:
    """The arguments for the random text API."""

    seed: str = ""
    words: int = 500

    def validate(self) -> None:
        """Validate this."""
        self.words = max(0, min(self.words, 2000))


WORDS = tuple(
    (DIR / "words.txt")
    .read_text(encoding="UTF-8")
    .replace("sozial", "asozial")
    .replace("Sozial", "Asozial")
    .splitlines()
)

HTML: str = (
    "<!DOCTYPE html>"
    '<html lang="de">'
    "<head>"
    "<title>{title}</title>"
    "</head>"
    "<body>"
    "<h1>{title}</h1>"
    "{text}"
    "</body>"
    "</html>"
)
HTML_BLOCKS: tuple[str, ...] = (
    "<blockquote>{p}</blockquote>",
    "<code>{p}</code>",
    "<pre>{p}</pre>",
)


def generate_random_word(random: Random, title: bool = False) -> str:
    """Generate a random word."""
    word = random.choice(WORDS)
    return word.title() if title else word


def generate_random_text(random: Random, length: int, end: str = "") -> str:
    """Generate a random text."""
    text: list[str] = []

    title_next = True

    for _ in range(length - 1):
        text.append(generate_random_word(random, title=title_next))
        add_punctuation = not random.randrange(9) and not title_next
        title_next = False
        if add_punctuation:
            text.append(random.choice(".?!.?!,;"))
            if text[-1] not in ",;":
                title_next = True
                if not random.randrange(4):
                    text.append("\n\n")
                    continue
        text.append(" ")

    text.extend((generate_random_word(random, title=title_next), end))

    return "".join(text)


class RandomText(APIRequestHandler):
    """The request handler for the random text API."""

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        "text/plain",
        "text/html",
    )

    @parse_args(type_=Arguments, validation_method="validate")
    async def get(self, args: Arguments, *, head: bool = False) -> None:
        """Handle GET requests to the random text API."""
        # pylint: disable=unused-argument
        random = Random(args.seed or None)
        text = generate_random_text(random, args.words, end=".")
        if self.content_type == "text/plain":
            return await self.finish(text)

        title = regex.sub(
            r"[.?!,;]*\s+",
            " ",
            generate_random_text(random, random.randint(2, 4)),
        )
        text = "".join(
            (
                f"<p>{paragraph}</p>"
                if random.randrange(3)
                else random.choice(HTML_BLOCKS).format(p=paragraph)
            )
            for paragraph in text.split("\n\n")
        )
        return await self.finish(HTML.format(title=title, text=text))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A module providing special auth tokens."""

from __future__ import annotations

import hmac
import math
from base64 import b64decode, b64encode
from datetime import datetime
from hashlib import blake2b
from typing import ClassVar, Literal, NamedTuple, TypeAlias, TypeGuard, get_args

from .utils import Permission

TokenVersion: TypeAlias = Literal["0"]
SUPPORTED_TOKEN_VERSIONS: tuple[TokenVersion, ...] = get_args(TokenVersion)


class ParseResult(NamedTuple):
    """The class representing a token."""

    token: str
    permissions: Permission
    valid_until: datetime
    salt: bytes


class InvalidTokenError(Exception):
    """Exception thrown for invalid or expired tokens."""


class InvalidTokenVersionError(InvalidTokenError):
    """Exception thrown when the token has an invalid version."""

    SUPPORTED_TOKEN_VERSIONS: ClassVar = SUPPORTED_TOKEN_VERSIONS


def is_supported_version(version: str) -> TypeGuard[TokenVersion]:
    """Check whether the argument is a supported token version."""
    return version in SUPPORTED_TOKEN_VERSIONS


def _split_token(token: str) -> tuple[TokenVersion, str]:
    """Split a token into version and the body of the token."""
    if not token:
        raise InvalidTokenError()

    version = token[0]
    if is_supported_version(version):
        return version, token[1:]

    raise InvalidTokenVersionError()


def parse_token(  # pylint: disable=inconsistent-return-statements
    token: str,
    *,
    secret: bytes | str,
    verify_time: bool = True,
) -> ParseResult:
    """Parse an auth token."""
    secret = secret.encode("UTF-8") if isinstance(secret, str) else secret
    version, token_body = _split_token(token)
    try:
        if version == "0":
            return _parse_token_v0(token_body, secret, verify_time=verify_time)
    except InvalidTokenError:
        raise
    except Exception as exc:
        raise InvalidTokenError from exc


def create_token(  # pylint: disable=too-many-arguments
    permissions: Permission,
    *,
    secret: bytes | str,
    duration: int,
    start: None | datetime = None,
    salt: None | bytes | str = None,
    version: TokenVersion = SUPPORTED_TOKEN_VERSIONS[-1],
) -> ParseResult:
    """Create an auth token."""
    secret = secret.encode("UTF-8") if isinstance(secret, str) else secret
    start = datetime.now() if start is None else start
    salt = salt.encode("UTF-8") if isinstance(salt, str) else salt or b""
    token: str
    if version == "0":
        token = _create_token_body_v0(
            permissions, secret, duration, start, salt
        )

    return parse_token(version + token, secret=secret, verify_time=False)


def int_to_bytes(number: int, length: int, signed: bool = False) -> bytes:
    """Convert an int to bytes."""
    return number.to_bytes(length, "big", signed=signed)


def bytes_to_int(bytes_: bytes, signed: bool = False) -> int:
    """Convert an int to bytes."""
    return int.from_bytes(bytes_, "big", signed=signed)


def _parse_token_v0(
    token_body: str, secret: bytes, *, verify_time: bool = True
) -> ParseResult:
    """Parse an auth token of version 0."""
    data: bytes = b64decode(token_body)
    data, hash_ = data[:-48], data[-48:]
    if not hmac.compare_digest(hmac.digest(secret, data, "SHA3-384"), hash_):
        raise InvalidTokenError()
    data, start = data[:-5], bytes_to_int(data[-5:])
    data, duration = data[:-5], bytes_to_int(data[-5:])
    permissions, salt = bytes_to_int(data[:-6]), data[-6:]

    now = int(datetime.now().timestamp())
    if verify_time and (now < start or start + duration < now):
        raise InvalidTokenError()

    return ParseResult(
        "0" + token_body,
        Permission(permissions),
        datetime.fromtimestamp(start + duration),
        salt,
    )


def _create_token_body_v0(
    permissions: Permission,
    secret: bytes,
    duration: int,
    start: datetime,
    salt: bytes,
) -> str:
    """Create an auth token of version 0."""
    if not salt:
        salt = blake2b(
            int_to_bytes(int(start.timestamp() - duration), 5), digest_size=6
        ).digest()
    elif len(salt) < 6:
        salt = b"U" * (6 - len(salt)) + salt
    elif len(salt) > 6:
        salt = salt[:6]

    parts = (
        int_to_bytes(permissions, math.ceil(len(Permission) / 8)),
        salt,
        int_to_bytes(duration, 5),
        int_to_bytes(int(start.timestamp()), 5),
    )
    data: bytes = b"".join(parts)

    len_token = len(data) + 384 // 8
    if len_token % 3:
        data = int_to_bytes(0, 3 - (len_token % 3)) + data

    hash_ = hmac.digest(secret, data, "SHA3-384")
    return b64encode(data + hash_).decode("UTF-8")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Options in this module are configurable when accessing the website."""

from __future__ import annotations

import dataclasses
import typing
from abc import ABC
from collections.abc import Callable, Iterable
from functools import partial
from typing import Final, Generic, Literal, TypeVar, overload

from tornado.web import RequestHandler

from . import base_request_handler as brh  # pylint: disable=cyclic-import
from .themes import THEMES
from .utils import (
    BumpscosityValue,
    OpenMojiValue,
    bool_to_str,
    parse_bumpscosity,
    parse_openmoji_arg,
    str_to_bool,
)

T = TypeVar("T")
U = TypeVar("U")

type ColourScheme = Literal["light", "dark", "system", "random"]
COLOUR_SCHEMES: Final[tuple[ColourScheme, ...]] = typing.get_args(
    ColourScheme.__value__  # pylint: disable=no-member
)


@dataclasses.dataclass(frozen=True)
class Option(ABC, Generic[T]):
    """An option that can be configured when accessing the website."""

    name: str
    parse_from_string: Callable[[str, T], T]
    get_default_value: Callable[[brh.BaseRequestHandler], T]
    is_valid: Callable[[T], bool] = lambda _: True
    normalize_string: Callable[[str], str] = lambda value: value
    value_to_string: Callable[[T], str] = str
    httponly: bool = True

    @overload
    def __get__(  # noqa: D105
        self, obj: None, _: type[Options] | None = None, /  # noqa: W504
    ) -> Option[T]: ...

    @overload
    def __get__(self, obj: Options, _: type[Options] | None = None, /) -> T:
        """Get the value for this option."""

    def __get__(
        self,  # comment to make Flake8 happy
        obj: Options | None,
        _: type[Options] | None = None,
        /,
    ) -> T | Option[T]:
        """Get the value for this option."""
        if obj is None:
            return self
        return self.get_value(obj.request_handler)

    def __set__(self, obj: brh.BaseRequestHandler, value: object) -> None:
        """Make this read-only."""
        raise AttributeError()

    def _parse(
        self,
        *,
        body_argument: str | None,
        query_argument: str | None,
        cookie: str | None,
        default: T,
    ) -> T:
        """Parse the value from a string."""
        for val in (body_argument, query_argument, cookie):
            if not val:
                continue
            parsed = self.parse_from_string(self.normalize_string(val), default)
            # is True to catch the case where is_valid returns NotImplemented
            if self.is_valid(parsed) is True:
                return parsed
        return default

    def get_form_appendix(self, request_handler: brh.BaseRequestHandler) -> str:
        """Return the form appendix for this option."""
        if not self.option_in_arguments(request_handler):
            return ""
        return (
            f"<input class='hidden' name={self.name!r} "
            f"value={self.value_to_string(self.get_value(request_handler))!r}>"
        )

    def get_value(
        self,
        request_handler: brh.BaseRequestHandler,
        *,
        include_body_argument: bool = True,
        include_query_argument: bool = True,
        include_cookie: bool = True,
    ) -> T:
        """Get the value for this option."""
        return self._parse(
            body_argument=(
                request_handler.get_body_argument(self.name, None)
                if include_body_argument
                else None
            ),
            query_argument=(
                request_handler.get_query_argument(self.name, None)
                if include_query_argument
                else None
            ),
            cookie=(
                request_handler.get_cookie(self.name, None)
                if include_cookie
                else None
            ),
            default=self.get_default_value(request_handler),
        )

    def option_in_arguments(
        self, request_handler: brh.BaseRequestHandler
    ) -> bool:
        """Return whether the option is taken from the arguments."""
        return self.get_value(
            request_handler, include_cookie=False
        ) != self.get_default_value(request_handler)


def parse_int(value: str, default: int) -> int:
    """Parse the value from a string."""
    try:
        return int(value, base=0)
    except ValueError:
        return default


def parse_string(value: str, _: str) -> str:
    """Parse the value from a string."""
    return value


StringOption = partial(Option[str], parse_from_string=parse_string)
BoolOption = partial(
    Option[bool], parse_from_string=str_to_bool, value_to_string=bool_to_str
)
IntOption = partial(Option[int], parse_from_string=parse_int)


def false(_: RequestHandler) -> Literal[False]:
    """Return False."""
    return False


def true(_: RequestHandler) -> Literal[True]:
    """Return True."""
    return True


def is_cautious_user(handler: RequestHandler) -> bool:
    """Return if a user is cautious."""
    return handler.request.host_name.endswith((".onion", ".i2p"))


class Options:
    """Options for the website."""

    __slots__ = ("__request_handler",)

    theme: Option[str] = StringOption(
        name="theme",
        is_valid=THEMES.__contains__,
        get_default_value=lambda _: (
            "default"
            # pylint: disable-next=misplaced-comparison-constant
            if _.now.month != 4 or 2 <= _.now.day
            else "fun"
        ),
        normalize_string=lambda s: s.replace("-", "_").lower(),
    )
    scheme: Option[ColourScheme] = Option(
        name="scheme",
        is_valid=COLOUR_SCHEMES.__contains__,
        get_default_value=lambda _: "system",
        normalize_string=str.lower,
        parse_from_string=lambda val, _: typing.cast(ColourScheme, val),
        httponly=False,
    )
    compat: Option[bool] = BoolOption(name="compat", get_default_value=false)
    dynload: Option[bool] = BoolOption(name="dynload", get_default_value=false)
    effects: Option[bool] = BoolOption(name="effects", get_default_value=true)
    openmoji: Option[OpenMojiValue] = Option(
        name="openmoji",
        parse_from_string=parse_openmoji_arg,
        get_default_value=false,
    )
    no_3rd_party: Option[bool] = BoolOption(
        name="no_3rd_party",
        get_default_value=is_cautious_user,
    )
    ask_before_leaving: Option[bool] = BoolOption(
        name="ask_before_leaving",
        get_default_value=false,
    )
    bumpscosity: Option[BumpscosityValue] = Option(
        name="bumpscosity",
        get_default_value=lambda _: parse_bumpscosity(None),
        parse_from_string=lambda v, u: parse_bumpscosity(v),
    )

    def __init__(self, request_handler: brh.BaseRequestHandler) -> None:
        """Initialize the options."""
        self.__request_handler = request_handler

    def as_dict(
        self,
        *,
        include_body_argument: bool = True,
        include_query_argument: bool = True,
        include_cookie: bool = True,
    ) -> dict[str, object]:
        """Get all the options in a dictionary."""
        return {
            option.name: option.get_value(
                self.request_handler,
                include_body_argument=include_body_argument,
                include_query_argument=include_query_argument,
                include_cookie=include_cookie,
            )
            for option in self.iter_options()
        }

    def as_dict_with_str_values(
        self,
        *,
        include_body_argument: bool = True,
        include_query_argument: bool = True,
        include_cookie: bool = True,
    ) -> dict[str, str]:
        """Get all the options in a dictionary."""
        return {
            option.name: option.value_to_string(
                option.get_value(
                    self.request_handler,
                    include_body_argument=include_body_argument,
                    include_query_argument=include_query_argument,
                    include_cookie=include_cookie,
                )
            )
            for option in self.iter_options()
        }

    def get_form_appendix(self) -> str:
        """Get HTML to add to forms to keep important query args."""
        return "".join(
            option.get_form_appendix(self.request_handler)
            for option in self.iter_options()
        )

    def iter_option_names(self) -> Iterable[str]:
        """Get the names of all options."""
        for option in self.iter_options():
            yield option.name

    def iter_options(self) -> Iterable[Option[object]]:
        """Get all the options."""
        for name in dir(self):
            if name.startswith("_"):
                continue
            value = getattr(self.__class__, name)
            if isinstance(value, Option):
                yield value

    @property
    def request_handler(self) -> brh.BaseRequestHandler:
        """Return the request handler."""
        return self.__request_handler



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The dataclass representing a quote of the day."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import date, datetime, timezone  # noqa: F401
from email.utils import format_datetime
from typing import Any

from ..utils import WrongQuote


@dataclass(slots=True)
class QuoteOfTheDayData:
    """The class representing data for the quote of the day."""

    date: date
    wrong_quote: WrongQuote
    url_without_path: str

    def get_date_for_rss(self) -> str:
        """Get the date as specified in RFC 2822."""
        return format_datetime(
            datetime(
                year=self.date.year,
                month=self.date.month,
                day=self.date.day,
                tzinfo=timezone.utc,
            ),
            True,
        )

    def get_guid(self) -> str:
        """Get the GUID for the quote of the day."""
        return f"{self.date.isoformat()}_{self.wrong_quote.get_id_as_str()}"

    def get_quote_as_str(self, new_line_char: str = "\n") -> str:
        """Get the quote as a string with new line."""
        return (
            f"»{self.wrong_quote.quote}«{new_line_char}"
            f"- {self.wrong_quote.author}"
        )

    def get_quote_image_url(self) -> str:
        """Get the URL of the image of the quote."""
        return self.get_quote_url() + ".gif"

    def get_quote_url(self) -> str:
        """Get the URL of the quote."""
        return (
            f"{self.url_without_path}/zitate/{self.wrong_quote.get_id_as_str()}"
        )

    def get_title(self) -> str:
        """Get the title for the quote of the day."""
        return f"Das Zitat des Tages vom {self.date:%d. %m. %Y}"

    def to_json(self) -> dict[str, str | dict[str, Any]]:
        """Get the quote of the day as JSON."""
        return {
            "date": self.date.isoformat(),
            "wrong_quote": self.wrong_quote.to_json(),
            "url": self.get_quote_url(),
            "image": self.get_quote_image_url(),
        }



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# pylint: disable=protected-access

"""Patches that improve everything."""

from __future__ import annotations

import asyncio
import http.client
import json as stdlib_json  # pylint: disable=preferred-module
import logging
import os
import sys
from collections.abc import Callable
from configparser import RawConfigParser
from contextlib import suppress
from importlib import import_module
from pathlib import Path
from threading import Thread
from types import MethodType
from typing import Any
from urllib.parse import urlsplit

import certifi
import defusedxml  # type: ignore[import-untyped]
import jsonpickle  # type: ignore[import-untyped]
import orjson
import pycurl
import tornado.httputil
import yaml
from emoji import EMOJI_DATA
from pillow_jxl import JpegXLImagePlugin  # noqa: F401
from setproctitle import setthreadtitle
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.httputil import HTTPFile, HTTPHeaders, HTTPServerRequest
from tornado.log import gen_log
from tornado.web import GZipContentEncoding, RedirectHandler, RequestHandler

from .. import CA_BUNDLE_PATH, MEDIA_TYPES
from . import braille, json  # noqa: F401  # pylint: disable=reimported


def apply() -> None:
    """Improve."""
    patch_asyncio()
    patch_certifi()
    patch_configparser()
    patch_emoji()
    patch_http()
    patch_json()
    patch_jsonpickle()
    patch_threading()
    patch_xml()

    patch_tornado_418()
    patch_tornado_arguments()
    patch_tornado_gzip()
    patch_tornado_httpclient()
    patch_tornado_logs()
    patch_tornado_redirect()


def patch_asyncio() -> None:
    """Make stuff faster."""
    if os.environ.get("DISABLE_UVLOOP") not in {
        "y", "yes", "t", "true", "on", "1"  # fmt: skip
    }:
        with suppress(ModuleNotFoundError):
            asyncio.set_event_loop_policy(
                import_module("uvloop").EventLoopPolicy()
            )


def patch_certifi() -> None:
    """Make everything use our CA bundle."""
    certifi.where = lambda: CA_BUNDLE_PATH
    certifi.contents = lambda: Path(certifi.where()).read_text("ASCII")


def patch_configparser() -> None:
    """Make configparser funky."""
    RawConfigParser.BOOLEAN_STATES.update(  # type: ignore[attr-defined]
        {
            "sure": True,
            "nope": False,
            "accept": True,
            "reject": False,
            "enabled": True,
            "disabled": False,
        }
    )


def patch_emoji() -> None:
    """Add cool new emoji."""
    EMOJI_DATA["🐱\u200D💻"] = {
        "de": ":hacker_katze:",
        "en": ":hacker_cat:",
        "status": 2,
        "E": 1,
    }
    for de_name, en_name, rect in (
        ("rot", "red", "🟥"),
        ("blau", "blue", "🟦"),
        ("orang", "orange", "🟧"),
        ("gelb", "yellow", "🟨"),
        ("grün", "green", "🟩"),
        ("lilan", "purple", "🟪"),
        ("braun", "brown", "🟫"),
    ):
        EMOJI_DATA[f"🫙\u200D{rect}"] = {
            "de": f":{de_name}es_glas:",
            "en": f":{en_name}_jar:",
            "status": 2,
            "E": 14,
        }
        EMOJI_DATA[f"🏳\uFE0F\u200D{rect}"] = {
            "de": f":{de_name}e_flagge:",
            "en": f":{en_name}_flag:",
            "status": 2,
            "E": 11,
        }
        EMOJI_DATA[f"\u2691\uFE0F\u200D{rect}"] = {
            "de": f":tief{de_name}e_flagge:",
            "en": f":deep_{en_name}_flag:",
            "status": 2,
            "E": 11,
        }


def patch_http() -> None:
    """Add response code 420."""
    http.client.responses[420] = "Enhance Your Calm"


def patch_json() -> None:
    """Replace json with orjson."""
    if getattr(stdlib_json, "_omegajson", False) or sys.version_info < (3, 12):
        return
    stdlib_json.dumps = json.dumps
    stdlib_json.dump = json.dump  # type: ignore[assignment]
    stdlib_json.loads = json.loads  # type: ignore[assignment]
    stdlib_json.load = json.load


def patch_jsonpickle() -> None:
    """Make jsonpickle return bytes."""
    jsonpickle.load_backend("orjson")
    jsonpickle.set_preferred_backend("orjson")
    jsonpickle.enable_fallthrough(False)


def patch_threading() -> None:
    """Set thread names."""
    _bootstrap = Thread._bootstrap  # type: ignore[attr-defined]

    def bootstrap(self: Thread) -> None:
        with suppress(Exception):
            setthreadtitle(self.name)
        _bootstrap(self)

    Thread._bootstrap = bootstrap  # type: ignore[attr-defined]


def patch_tornado_418() -> None:
    """Add support for RFC 7168."""
    RequestHandler.SUPPORTED_METHODS += (
        "PROPFIND",
        "BREW",
        "WHEN",
    )
    _ = RequestHandler._unimplemented_method
    RequestHandler.propfind = _  # type: ignore[attr-defined]
    RequestHandler.brew = _  # type: ignore[attr-defined]
    RequestHandler.when = _  # type: ignore[attr-defined]


def patch_tornado_arguments() -> None:  # noqa: C901
    """Improve argument parsing."""
    # pylint: disable=too-complex

    def ensure_bytes(value: Any) -> bytes:
        """Return the value as bytes."""
        if isinstance(value, bool):
            return b"true" if value else b"false"
        if isinstance(value, bytes):
            return value
        return str(value).encode("UTF-8")

    def parse_body_arguments(
        content_type: str,
        body: bytes,
        arguments: dict[str, list[bytes]],
        files: dict[str, list[HTTPFile]],
        headers: None | HTTPHeaders = None,
        *,
        _: Callable[..., None] = tornado.httputil.parse_body_arguments,
    ) -> None:
        # pylint: disable=too-many-branches
        if content_type.startswith("application/json"):
            if headers and "Content-Encoding" in headers:
                gen_log.warning(
                    "Unsupported Content-Encoding: %s",
                    headers["Content-Encoding"],
                )
                return
            try:
                spam = orjson.loads(body)
            except Exception as exc:  # pylint: disable=broad-except
                gen_log.warning("Invalid JSON body: %s", exc)
            else:
                if not isinstance(spam, dict):
                    return
                for key, value in spam.items():
                    if value is not None:
                        arguments.setdefault(key, []).append(
                            ensure_bytes(value)
                        )
        elif content_type.startswith("application/yaml"):
            if headers and "Content-Encoding" in headers:
                gen_log.warning(
                    "Unsupported Content-Encoding: %s",
                    headers["Content-Encoding"],
                )
                return
            try:
                spam = yaml.safe_load(body)
            except Exception as exc:  # pylint: disable=broad-except
                gen_log.warning("Invalid YAML body: %s", exc)
            else:
                if not isinstance(spam, dict):
                    return
                for key, value in spam.items():
                    if value is not None:
                        arguments.setdefault(key, []).append(
                            ensure_bytes(value)
                        )
        else:
            _(content_type, body, arguments, files, headers)

    parse_body_arguments.__doc__ = tornado.httputil.parse_body_arguments.__doc__

    tornado.httputil.parse_body_arguments = parse_body_arguments


def patch_tornado_gzip() -> None:
    """Use gzip for more content types."""
    GZipContentEncoding.CONTENT_TYPES = {
        type for type, data in MEDIA_TYPES.items() if data.get("compressible")
    }


def patch_tornado_httpclient() -> None:  # fmt: off
    """Make requests quick."""
    BACON = 0x75800  # noqa: N806  # pylint: disable=invalid-name
    EGGS = 1 << 25  # noqa: N806  # pylint: disable=invalid-name

    AsyncHTTPClient.configure("tornado.curl_httpclient.CurlAsyncHTTPClient")

    def prepare_curl_callback(self: HTTPRequest, curl: pycurl.Curl) -> None:
        # pylint: disable=c-extension-no-member, useless-suppression
        if urlsplit(self.url).scheme == "https":  # noqa: SIM102
            if (ver := pycurl.version_info())[2] >= BACON and ver[4] & EGGS:
                curl.setopt(pycurl.HTTP_VERSION, pycurl.CURL_HTTP_VERSION_3)

    original_request_init = HTTPRequest.__init__

    def request_init(self: HTTPRequest, *args: Any, **kwargs: Any) -> None:
        if len(args) < 18:  # there are too many positional arguments here
            prepare_curl_method = MethodType(prepare_curl_callback, self)
            kwargs.setdefault("prepare_curl_callback", prepare_curl_method)
        original_request_init(self, *args, **kwargs)

    request_init.__doc__ = HTTPRequest.__init__.__doc__

    HTTPRequest.__init__ = request_init  # type: ignore[method-assign]


def patch_tornado_logs() -> None:
    """Anonymize Tornado logs."""
    # pylint: disable=import-outside-toplevel
    from ..utils.utils import SUS_PATHS, anonymize_ip

    RequestHandler._request_summary = (  # type: ignore[method-assign]
        lambda self: "%s %s (%s)"  # pylint: disable=consider-using-f-string
        % (
            self.request.method,
            self.request.uri,
            (
                self.request.remote_ip
                if self.request.path == "/robots.txt"
                or self.request.path.lower() in SUS_PATHS
                else anonymize_ip(self.request.remote_ip, ignore_invalid=True)
            ),
        )
    )

    HTTPServerRequest.__repr__ = (  # type: ignore[method-assign]
        lambda self: "%s(%s)"  # pylint: disable=consider-using-f-string
        % (
            self.__class__.__name__,
            ", ".join(
                [
                    "%s=%r"  # pylint: disable=consider-using-f-string
                    % (
                        n,
                        getattr(self, n),
                    )
                    for n in ("protocol", "host", "method", "uri", "version")
                ]
            ),
        )
    )


def patch_tornado_redirect() -> None:
    """Use modern redirect codes and support HEAD requests."""

    def redirect(
        self: RequestHandler,
        url: str,
        permanent: bool = False,
        status: None | int = None,
    ) -> None:
        if url == self.request.full_url():
            logging.getLogger(
                f"{self.__class__.__module__}.{self.__class__.__qualname__}"
            ).critical("Infinite redirect to %r detected", url)
        if self._headers_written:
            # pylint: disable=broad-exception-raised
            raise Exception("Cannot redirect after headers have been written")
        if status is None:
            status = 308 if permanent else 307
        else:
            assert isinstance(status, int) and 300 <= status <= 399  # type: ignore[redundant-expr]  # noqa: B950
        self.set_status(status)
        self.set_header("Location", url)
        self.finish()  # type: ignore[unused-awaitable]

    if RequestHandler.redirect.__doc__:
        # fmt: off
        redirect.__doc__ = (
            RequestHandler.redirect.__doc__
            .replace("301", "308")
            .replace("302", "307")
        )
        # fmt: on

    RequestHandler.redirect = redirect  # type: ignore[method-assign]

    RedirectHandler.head = RedirectHandler.get


def patch_xml() -> None:
    """Make XML safer."""
    defusedxml.defuse_stdlib()
    defusedxml.xmlrpc.monkey_patch()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module used for easy and simple searching in O(n) complexity."""

from __future__ import annotations

import dataclasses
from collections.abc import Callable, Iterable, Iterator, Sequence
from typing import Generic, NoReturn, TypeVar

import regex as re
from typed_stream import Stream

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")


class Query:
    """Class representing a query."""

    _data: tuple[str, Sequence[str], int]
    __slots__ = ("_data",)

    def __bool__(self) -> bool:
        """Return False if this query matches everything."""
        return bool(self.words)

    def __hash__(self) -> int:
        """Hash this."""
        return hash(self.query)

    def __init__(self, query: str) -> None:
        """Initialize this."""
        if hasattr(self, "_data"):
            raise ValueError("Already initialized.")
        query = query.lower()
        # pylint: disable=bad-builtin
        words = tuple(filter(None, re.split(r"\W+", query)))
        words_len = sum(len(word) for word in words)
        object.__setattr__(self, "_data", (query, words, words_len))

    def __reduce__(self) -> tuple[type[Query], tuple[str]]:
        """Reduce this object."""
        return Query, (self.query,)

    def __repr__(self) -> str:
        """Return a string representation of self."""
        return f"Query({self.query!r})"

    def __setattr__(self, key: object, value: object) -> NoReturn:
        """Raise an AttributeError."""
        raise AttributeError("Cannot modify Query.")

    @property
    def query(self) -> str:
        """The original query."""
        return self._data[0]

    def score(self, field_values: tuple[str, ...]) -> float:
        """Field values needs to be a tuple of lower cased strings."""
        if any(self.query in val for val in field_values):
            return 1.0
        return sum(
            (
                sum(word in value for value in field_values)
                * (len(word) / self.words_len)
            )
            for word in self.words
        ) / len(field_values)

    @property
    def words(self) -> Sequence[str]:
        """The words in the query."""
        return self._data[1]

    @property
    def words_len(self) -> int:
        """Return sum(len(word) for word in self.words)."""
        return self._data[2]


@dataclasses.dataclass(frozen=True, slots=True, order=True)
class ScoredValue(Generic[T]):
    """Value with score."""

    score: float
    value: T


class DataProvider(Generic[T, U]):
    """Provide Data."""

    __slots__ = ("_data", "_key", "_convert")

    _data: Iterable[T] | Callable[[], Iterable[T]]
    _key: Callable[[T], str | tuple[str, ...]]
    _convert: Callable[[T], U]

    def __init__(
        self,
        data: Iterable[T] | Callable[[], Iterable[T]],
        key: Callable[[T], str | tuple[str, ...]],
        convert: Callable[[T], U],
    ) -> None:
        """Initialize this."""
        self._data = data
        self._key = key
        self._convert = convert

    def _value_to_fields(self, value: T) -> tuple[str, ...]:
        """Convert a value to a tuple of strings."""
        return (
            (cpv.lower(),)
            if isinstance(cpv := self._key(value), str)
            else tuple(map(str.lower, cpv))  # pylint: disable=bad-builtin
        )

    @property
    def data(self) -> Iterable[T]:
        """Return the data."""
        return self._data if isinstance(self._data, Iterable) else self._data()

    def search(
        self, query: Query, excl_min_score: float = 0.0
    ) -> Iterator[ScoredValue[U]]:
        """Search this."""
        for value in self.data:
            score = query.score(self._value_to_fields(value))
            if score > excl_min_score:
                yield ScoredValue(score, self._convert(value))


def search(
    query: Query,
    *providers: DataProvider[object, T],
    excl_min_score: float = 0.0,
) -> list[ScoredValue[T]]:
    """Search through data."""
    return sorted(
        Stream(providers).flat_map(lambda x: x.search(query, excl_min_score)),
        key=lambda sv: sv.score,
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Share page to share quotes on multiple services."""

from __future__ import annotations

from urllib.parse import quote

from .utils import QuoteReadyCheckHandler, get_wrong_quote


class ShareQuote(QuoteReadyCheckHandler):
    """Request handler for the share page."""

    async def get(
        self, quote_id: str, author_id: str, *, head: bool = False
    ) -> None:
        """Handle GET requests to the share page."""
        wrong_quote = await get_wrong_quote(int(quote_id), int(author_id))
        if head:
            return
        text = f"»{wrong_quote.quote.quote}«\n- {wrong_quote.author.name}"
        await self.render(
            "pages/quotes/share.html",
            text=quote(text, safe=""),
            u_text=text,
            u_url=(url := self.fix_url(f"/zitate/{quote_id}-{author_id}")),
            quote_url=quote(url, safe=""),
            wrong_quote=str(wrong_quote),
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The backdoor API of the website."""

from __future__ import annotations

from ..utils.utils import ModuleInfo
from .backdoor import Backdoor


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=((r"/api/backdoor/(eval|exec)", Backdoor),),
        name="Backdoor",
        description="🚪",
        aliases=("/api/hintertür", "/api/hintertuer"),
        hidden=True,
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the permission system of an-website."""

from __future__ import annotations

import socket

from . import (  # noqa: F401  # pylint: disable=unused-import
    FetchCallable,
    app,
    assert_valid_response,
    fetch,
)
from .test_backdoor import request_and_parse


async def test_permissions(
    fetch: FetchCallable,  # noqa: F811
    http_server_port: tuple[socket.socket, int],
) -> None:
    """Test stuff with permissions."""
    for key, headers in (
        (
            "",
            {
                "X-Permission-Backdoor": "nope",
                "X-Permission-Ratelimits": "nope",
                "X-Permission-Update": "nope",
                "X-Permission-Traceback": "nope",
            },
        ),
        (
            "s0",
            {
                "X-Permission-Backdoor": "nope",
                "X-Permission-Ratelimits": "nope",
                "X-Permission-Update": "nope",
                "X-Permission-Traceback": "nope",
            },
        ),
        (
            "s1",
            {
                "X-Permission-Backdoor": "nope",
                "X-Permission-Ratelimits": "sure",
                "X-Permission-Update": "nope",
                "X-Permission-Traceback": "nope",
            },
        ),
        (
            "s2",
            {
                "X-Permission-Backdoor": "nope",
                "X-Permission-Ratelimits": "nope",
                "X-Permission-Update": "nope",
                "X-Permission-Traceback": "sure",
            },
        ),
        (
            "s3",
            {
                "X-Permission-Backdoor": "nope",
                "X-Permission-Ratelimits": "sure",
                "X-Permission-Update": "nope",
                "X-Permission-Traceback": "sure",
            },
        ),
        (
            "s4",
            {
                "X-Permission-Backdoor": "sure",
                "X-Permission-Ratelimits": "nope",
                "X-Permission-Update": "nope",
                "X-Permission-Traceback": "nope",
            },
        ),
        (
            "123qweQWE!@#000000000",
            {
                "X-Permission-Backdoor": "sure",
                "X-Permission-Ratelimits": "sure",
                "X-Permission-Update": "sure",
                "X-Permission-Traceback": "sure",
            },
        ),
    ):
        assert_valid_response(
            await fetch("/api/ping", headers={"Authorization": key}),
            "text/plain;charset=utf-8",
            {200},
            headers=headers,
        )
        assert_valid_response(
            await fetch(f"/api/ping?key={key.replace('#', '%23')}"),
            "text/plain;charset=utf-8",
            {200},
            headers=headers,
        )

    assert_valid_response(
        await fetch("/api/ping?key=s3", headers={"Authorization": "s4"}),
        "text/plain;charset=utf-8",
        {200},
        headers={
            "X-Permission-Backdoor": "sure",
            "X-Permission-Ratelimits": "sure",
            "X-Permission-Update": "nope",
            "X-Permission-Traceback": "sure",
        },
    )


async def test_permissions_with_backdoor(
    fetch: FetchCallable,  # noqa: F811
) -> None:
    """Test permissions with the backdoor."""
    response = await request_and_parse(
        fetch, "app.settings['TRUSTED_API_SECRETS'].get('')"
    )
    assert isinstance(response, dict)
    assert response["success"]
    assert not response["output"]
    assert response["result"] is None

    response = await request_and_parse(
        fetch, "len(app.settings['TRUSTED_API_SECRETS'])"
    )
    assert isinstance(response, dict)
    assert response["success"]
    assert not response["output"]
    assert response["result"] == ("6", 6)

    response = await request_and_parse(
        fetch,
        (
            "from an_website.utils.utils import"
            " Permission\nprint(self.is_authorized(Permission.TRACEBACK |"
            " Permission.BACKDOOR))"
        ),
        auth_key="s4",
        mode="exec?key=s3",  # type: ignore[arg-type]
    )
    assert isinstance(response, dict)
    assert response["success"]
    assert response["output"] == "True\n"
    assert response["result"] is None

    response = await request_and_parse(
        fetch,
        (
            "from an_website.utils.utils import Permission\n"
            "print(self.is_authorized(Permission.RATELIMITS))"
        ),
        mode="exec",
    )
    assert isinstance(response, dict)
    assert response["success"]
    assert response["output"] == "True\n"
    assert response["result"] is None

    assert_valid_response(
        await fetch(
            "/api/backdoor/exec",
            headers={"Authorization": "unknown"},
            method="POST",
            body="42",
        ),
        None,
        {401},  # unauthenticated request
    )
    assert_valid_response(
        await fetch(
            "/api/backdoor/exec",
            headers={"Authorization": "s3"},
            method="POST",
            body="42",
        ),
        None,
        {403},  # unauthorized request
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the uptime page."""

from __future__ import annotations

from an_website import UPTIME
from an_website.uptime import uptime
from an_website.utils import utils


def test_calculate_uptime() -> None:
    """Test the calculation of the uptime."""
    assert UPTIME.get() < UPTIME.get()
    assert UPTIME.get() + 1 > UPTIME.get()

    assert utils.time_to_str(UPTIME.get())


def test_get_availability_dict() -> None:
    """Test the get_availability_dict function."""
    assert uptime.get_availability_dict(10, 90) == {
        "up": 10,
        "down": 90,
        "total": 100,
        "percentage": 10.0,
    }
    assert uptime.get_availability_dict(90, 10) == {
        "up": 90,
        "down": 10,
        "total": 100,
        "percentage": 90.0,
    }
    assert uptime.get_availability_dict(50, 50) == {
        "up": 50,
        "down": 50,
        "total": 100,
        "percentage": 50.0,
    }
    assert uptime.get_availability_dict(42, 0) == {
        "up": 42,
        "down": 0,
        "total": 42,
        "percentage": 100.0,
    }
    assert uptime.get_availability_dict(0, 0) == {
        "up": 0,
        "down": 0,
        "total": 0,
        "percentage": None,
    }


if __name__ == "__main__":
    test_calculate_uptime()
    test_get_availability_dict()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# pylint: disable=missing-function-docstring, unused-argument
# fmt: off

"""⠗⡰⠓"""

from __future__ import annotations

from codecs import CodecInfo, register


def encode(data: str, errors: str = "strict") -> tuple[bytes, int]:
    output = bytearray()
    for pos, char in enumerate(data):
        byte = ord(char) - 0x2800
        if not 0 <= byte <= 255:
            if errors == "ignore":
                continue
            raise UnicodeEncodeError(
                "braille",
                char,
                pos,
                pos + 1,
                "ordinal not in range(0x2800, 0x2900)",
            )
        output.append(byte)
    return bytes(output), len(data)


def decode(data: bytes, errors: str = "strict") -> tuple[str, int]:
    return "".join(chr(0x2800 + byte) for byte in data), len(data)


def morb() -> CodecInfo:
    """⡉⡔⠧⡓⠠⡍⡏⡒⡂⡉⡎⠧⠠⡔⡉⡍⡅"""
    return CodecInfo(
        name="braille",
        encode=encode,
        decode=decode,  # type: ignore[arg-type]
    )


register(lambda codec: morb() if codec == "braille" else None)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
A page with all the JavaScript files and their licenses.

This is used for LibreJS to make sure the extension knows the licenses.
This isn't important, as the JS files should contain the licenses themselves.

See: https://www.gnu.org/software/librejs/free-your-javascript.html#step3
and https://www.gnu.org/licenses/javascript-labels.html
"""

from __future__ import annotations

import logging
from collections.abc import Mapping
from functools import cache
from typing import Final

from .. import STATIC_DIR, pytest_is_running
from ..utils.fix_static_path_impl import recurse_directory
from ..utils.request_handler import HTMLRequestHandler
from ..utils.utils import ModuleInfo

LOGGER: Final = logging.getLogger(__name__)


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=((r"/js-lizenzen", JSLicenses),),
        name="JavaScript-Lizenzen",
        description=(
            "Informationen über die Lizenzen der JavaScript-Dateien "
            "auf dieser Seite"
        ),
        path="/js-lizenzen",
        aliases=("/js-licenses",),
        keywords=("JavaScript", "License", "Lizenz"),
    )


LICENSE_COMMENT_START: Final = "// @license "
LICENSES: Final[Mapping[str, str]] = {
    "magnet:?xt=urn:btih:0b31508aeb0634b347b8270c7bee4d411b5d4109&dn=agpl-3.0.txt": (
        "https://www.gnu.org/licenses/agpl-3.0.html"
    ),
    "magnet:?xt=urn:btih:d3d9a9a6595521f9666a5e94cc830dab83b65699&dn=expat.txt": (
        "https://www.jclark.com/xml/copying.txt"
    ),
}


@cache
def get_js_filenames_and_licenses() -> list[tuple[str, str, str]]:
    """
    Get the names of the JS files in this project.

    Returns a list of tuples with filename, license and license URL.
    """
    js_files_dir = STATIC_DIR / "js"
    licenses_list: list[tuple[str, str, str]] = []
    for filename in recurse_directory(
        js_files_dir, lambda path: path.name.endswith(".js")
    ):
        path = js_files_dir / filename
        if not path.is_file():
            continue
        with path.open(encoding="UTF-8") as file:
            license_line = file.readline()
        if LICENSE_COMMENT_START not in license_line:
            LOGGER.critical("%s has no license comment", filename)
            if pytest_is_running():
                raise AssertionError(f"Could not get license of: {filename}")
            continue
        magnet, name = (
            license_line[
                license_line.index(LICENSE_COMMENT_START)
                + len(LICENSE_COMMENT_START) :
            ]
            .strip()
            .split(" ")
        )
        licenses_list.append((filename, name.strip(), LICENSES[magnet.strip()]))
    return licenses_list


class JSLicenses(HTMLRequestHandler):
    """The request handler for the JS licenses page."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the JS licenses page."""
        if head:
            return
        await self.render(
            "pages/js_licenses.html", js_files=get_js_filenames_and_licenses()
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Info page to show information about authors and quotes."""

from __future__ import annotations

import logging
from datetime import datetime, timedelta, timezone
from typing import Final, cast
from urllib.parse import quote as quote_url

import orjson as json
import regex
from tornado.httpclient import AsyncHTTPClient

from .. import CA_BUNDLE_PATH, EVENT_REDIS
from ..utils.request_handler import HTMLRequestHandler
from .utils import get_author_by_id, get_quote_by_id, get_wrong_quotes

LOGGER: Final = logging.getLogger(__name__)


class QuotesInfoPage(HTMLRequestHandler):
    """The request handler used for the info page."""

    RATELIMIT_GET_LIMIT = 30

    async def get(self, id_str: str, *, head: bool = False) -> None:
        """Handle GET requests to the quote info page."""
        quote_id: int = int(id_str)
        quote = await get_quote_by_id(quote_id)
        if head:
            return
        wqs = get_wrong_quotes(lambda wq: wq.quote_id == quote_id, sort=True)
        await self.render(
            "pages/quotes/quote_info.html",
            quote=quote,
            wrong_quotes=wqs,
            title="Zitat-Informationen",
            short_title="Zitat-Info",
            type="Zitat",
            id=quote_id,
            text=str(quote),
            description=f"Falsch zugeordnete Zitate mit „{quote}“ als Zitat.",
            create_kwargs={"quote": quote_id},
        )


WIKI_API_DE: Final[str] = "https://de.wikipedia.org/w/api.php"
WIKI_API_EN: Final[str] = "https://en.wikipedia.org/w/api.php"


async def search_wikipedia(
    query: str, api: str = WIKI_API_DE
) -> None | tuple[str, None | str, datetime]:
    """
    Search Wikipedia to get information about the query.

    Return a tuple with the URL and the content.
    """
    if not query:
        return None
    # try to get the info from Wikipedia
    response = await AsyncHTTPClient().fetch(
        (
            f"{api}?action=opensearch&namespace=0&profile=normal&"
            f"search={quote_url(query)}&limit=1&redirects=resolve&format=json"
        ),
        ca_certs=CA_BUNDLE_PATH,
    )
    response_json = json.loads(response.body)
    if not response_json[1]:
        if api == WIKI_API_DE:
            return await search_wikipedia(query, WIKI_API_EN)
        return None  # nothing found
    page_name = response_json[1][0]
    # get the URL of the content & replace "," with "%2C"
    url = str(response_json[3][0]).replace(",", "%2C")

    return (
        url,
        await get_wikipedia_page_content(page_name, api),
        datetime.now(timezone.utc),
    )


async def get_wikipedia_page_content(
    page_name: str, api: str = WIKI_API_DE
) -> None | str:
    """Get content from a Wikipedia page and return it."""
    response = await AsyncHTTPClient().fetch(
        (
            f"{api}?action=query&prop=extracts&exsectionformat=plain&exintro&"
            f"titles={quote_url(page_name)}&explaintext&format=json&exsentences=5"
        ),
        ca_certs=CA_BUNDLE_PATH,
    )
    response_json = json.loads(response.body)
    if "query" not in response_json or "pages" not in response_json["query"]:
        return None
    pages: dict[str, str] = response_json["query"]["pages"]
    page = cast(dict[str, str], tuple(pages.values())[0])
    if "extract" not in page:
        return None
    return page["extract"]


def fix_author_for_wikipedia_search(author: str) -> str:
    """
    Fix author for Wikipedia search.

    This tries to reduce common problems with authors.
    So that we can show more information.
    """
    author = regex.sub(r"\s+", " ", author)
    author = regex.sub(r"\s*\(.*\)", "", author)
    author = regex.sub(r"\s*Werbespruch$", "", author, regex.IGNORECASE)
    author = regex.sub(r"\s*Werbung$", "", author, regex.IGNORECASE)
    author = regex.sub(r"^nach\s*", "", author, regex.IGNORECASE)
    author = regex.sub(r"^Ein\s+", "", author, regex.IGNORECASE)
    return author


# time to live in seconds (1 month)
AUTHOR_INFO_NEW_TTL: Final[int] = 60 * 60 * 24 * 30


class AuthorsInfoPage(HTMLRequestHandler):
    """The request handler used for the info page."""

    RATELIMIT_GET_LIMIT = 5

    async def get(self, id_str: str, *, head: bool = False) -> None:
        """Handle GET requests to the author info page."""
        author_id: int = int(id_str)
        author = await get_author_by_id(author_id)
        if head:
            return
        if author.info is None:
            result = None
            fixed_author_name = fix_author_for_wikipedia_search(author.name)
            if EVENT_REDIS.is_set():
                # try to get the info from Redis
                result = await self.redis.get(
                    self.get_redis_info_key(fixed_author_name)
                )
            if result and (len(info := result.split("|", maxsplit=1)) > 1):
                remaining_ttl = await self.redis.ttl(
                    self.get_redis_info_key(fixed_author_name)
                )
                creation_date = datetime.now(tz=timezone.utc) - timedelta(
                    seconds=AUTHOR_INFO_NEW_TTL - remaining_ttl
                )
                if len(info) == 1:
                    author.info = (info[0], None, creation_date)
                else:
                    author.info = (info[0], info[1], creation_date)
            else:
                author.info = await search_wikipedia(fixed_author_name)
                if author.info is None or author.info[1] is None:
                    # nothing found
                    LOGGER.info("No information found about %s", repr(author))
                elif EVENT_REDIS.is_set():
                    await self.redis.setex(
                        self.get_redis_info_key(fixed_author_name),
                        AUTHOR_INFO_NEW_TTL,
                        # value to save (the author info)
                        # type is ignored, because author.info[1] is not None
                        "|".join(author.info[0:2]),  # type: ignore[arg-type]
                    )

        wqs = get_wrong_quotes(
            lambda wq: wq.author_id == author_id,
            sort=True,
        )

        await self.render(
            "pages/quotes/author_info.html",
            author=author,
            wrong_quotes=wqs,
            title="Autor-Informationen",
            short_title="Autor-Info",
            type="Autor",
            id=author_id,
            text=str(author),
            description=f"Falsch zugeordnete Zitate mit „{author}“ als Autor.",
            create_kwargs={"author": author_id},
        )

    def get_redis_info_key(self, author_name: str) -> str:
        """Get the key to save the author info with Redis."""
        return f"{self.redis_prefix}:quote-author-info:{author_name}"



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The ping API of the website."""

from __future__ import annotations

from typing import ClassVar

from ..utils.request_handler import APIRequestHandler
from ..utils.utils import ModuleInfo


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=((r"/api/ping", PingPong),),
        name="Ping Pong",
        description="🏓",
        path="/api/ping",
        hidden=True,
    )


class PingPong(APIRequestHandler):
    """The request handler for the ping API."""

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        "text/plain",
        *APIRequestHandler.POSSIBLE_CONTENT_TYPES,
    )

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the ping API."""
        # pylint: disable=unused-argument
        if self.content_type == "text/plain":
            await self.finish("🏓")
        else:
            await self.finish({"success": True})



#!/usr/bin/env python3

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The client for the backdoor API of the website."""

from __future__ import annotations

import ast
import asyncio
import http.client
import io
import os
import pickle  # nosec: B403
import pydoc
import re  # pylint: disable=preferred-module
import socket
import sys
import time
import traceback
import uuid
from base64 import b64encode
from collections.abc import Callable, Iterable, MutableMapping
from contextlib import suppress
from importlib import import_module
from textwrap import dedent
from types import EllipsisType
from typing import Any, Required, TypeAlias, TypedDict, cast
from urllib.parse import SplitResult, quote, quote_plus, urlsplit

with suppress(ModuleNotFoundError):
    # pylint: disable=shadowed-import
    import dill as pickle  # type: ignore[import-untyped, no-redef]  # noqa: F811, B950  # nosec: B403

try:
    import hy  # type: ignore[import-untyped]
except ModuleNotFoundError:
    hy = None  # pylint: disable=invalid-name

try:
    import idna
except ModuleNotFoundError:
    idna = None  # type: ignore[assignment]  # pylint: disable=invalid-name

try:
    import socks  # type: ignore[import-untyped]
except ModuleNotFoundError:
    socks = None  # pylint: disable=invalid-name

if os.environ.get("DISABLE_UVLOOP") not in {"y", "yes", "t", "true", "on", "1"}:
    with suppress(ModuleNotFoundError):
        asyncio.set_event_loop_policy(import_module("uvloop").EventLoopPolicy())

E = eval(  # pylint: disable=eval-used  # nosec: B307
    "eval(repr((_:=[],_.append(_))[0]))[0][0]"
)

ErrorTuple: TypeAlias = tuple[int, str]

FLUFL = True


class Response(TypedDict):  # noqa: D101
    # pylint: disable=missing-class-docstring
    success: bool | EllipsisType
    output: None | str
    result: None | tuple[str, None | bytes] | SystemExit


class Proxy(TypedDict, total=False):  # noqa: D101
    # pylint: disable=missing-class-docstring
    type: Required[int]
    addr: Required[str]
    port: int
    rdns: bool
    username: str
    password: str


async def create_socket(
    addr: str,
    port: int | str,
    proxy: None | Proxy,
) -> socket.socket:
    """Create a socket (optionally with a proxy)."""
    # pylint: disable=too-complex
    if proxy is not None and not socks:
        raise NotImplementedError("PySocks is required for proxy support")
    loop = asyncio.get_running_loop()
    address_infos = await loop.getaddrinfo(
        addr,
        port,
        # PySocks doesn't support AF_INET6
        family=0 if proxy is None else socket.AF_INET,
        type=socket.SOCK_STREAM,
    )
    if not address_infos:
        raise OSError("getaddrinfo() returned empty list")
    exceptions = []
    for address_info in address_infos:
        sock = None
        try:
            sock = cast(
                socket.socket,
                (
                    socks.socksocket(address_info[0])
                    if proxy is not None
                    else socket.socket(address_info[0])
                ),
            )
            sock.setblocking(False)
            if proxy is not None:
                sock.set_proxy(  # type: ignore[attr-defined]
                    proxy["type"],
                    proxy["addr"],
                    proxy.get("port"),
                    proxy.get("rdns", True),
                    proxy.get("username"),
                    proxy.get("password"),
                )
            await loop.sock_connect(sock, address_info[4])
            return sock
        except OSError as exc:
            if sock is not None:
                sock.close()
            exceptions.append(exc)
            continue
        except BaseException:
            if sock is not None:
                sock.close()
            raise
    if len(exceptions) == 1:
        raise exceptions[0]
    model = str(exceptions[0])
    if all(str(exc) == model for exc in exceptions):
        raise exceptions[0]
    raise OSError(
        f"Multiple exceptions: {', '.join(str(exc) for exc in exceptions)}"
    )


async def request(  # noqa: C901
    method: str,
    url: str | SplitResult,
    headers: None | MutableMapping[str, str] = None,
    body: None | bytes | bytearray | Iterable[bytes | bytearray] | str = None,
    proxy: None | Proxy = None,
) -> tuple[int, dict[str, str], bytes]:
    """Insanely awesome HTTP client."""
    # pylint: disable=line-too-long, too-complex
    # pylint: disable=too-many-branches, too-many-locals, while-used
    if isinstance(url, str):
        url = urlsplit(url)
    if url.scheme not in {"", "http", "https"}:
        raise ValueError(f"Unsupported scheme: {url.scheme}")
    if not url.hostname:
        raise ValueError("URL has no hostname")
    if headers is None:
        headers = {}
    if isinstance(body, str):
        body = body.encode("UTF-8")
    if isinstance(body, memoryview):
        body = body.tobytes()
    if isinstance(body, Iterable) and not isinstance(body, (bytes, bytearray)):
        body = b"".join(body)  # type: ignore[arg-type]
    https = url.scheme == "https"
    header_names = [x.strip().title() for x in headers.keys()]
    if "Host" not in header_names:
        host: None | str = None
        if idna:  # type: ignore[truthy-bool]
            with suppress(idna.core.InvalidCodepoint):
                host = idna.encode(url.hostname).decode("ASCII")
                host = f"{host}:{url.port}" if url.port else host
        if not host:
            host = url.netloc.encode("IDNA").decode("ASCII")
        headers["Host"] = host
    if body and "Content-Length" not in header_names:
        headers["Content-Length"] = str(len(body))
    sock = await create_socket(
        url.hostname,
        url.port or ("https" if https else "http"),
        proxy,
    )
    reader, writer = await asyncio.open_connection(
        sock=sock,
        ssl=https,
        server_hostname=url.hostname if https else None,
    )
    writer.write(
        (
            method
            + " "
            + (quote(url.path) or "/")
            + ("?" + quote_plus(url.query) if url.query else "")
            + " HTTP/1.0\r\n"
        ).encode("ASCII")
        + "\r\n".join(
            [f"{key}:{value}" for key, value in headers.items()] + [""]
        ).encode("LATIN-1")
        + b"\r\n"
    )
    if body:
        writer.write(body)
    await writer.drain()
    del headers
    e, data = E, b""
    while chunk := await reader.read():
        if b"\r\n\r\n" in (data := data + chunk) and e is E:
            e, data = data.split(b"\r\n\r\n", 1)
            status, o = re.match(r"HTTP/.+? (\d+).*?\r\n(.*)", e.decode("LATIN-1"), 24).groups()  # type: ignore[union-attr]  # noqa: B950
            headers = dict(re.match(r"([^\s]+):\s*(.+?)\s*$", x, 24).groups() for x in o.split("\r\n"))  # type: ignore[union-attr, misc]  # noqa: B950
    writer.close()
    await writer.wait_closed()
    if "status" not in locals():
        raise AssertionError("No HTTP response received")
    # pylint: disable-next=possibly-used-before-assignment
    return int(status), headers, data  # type: ignore[possibly-undefined]


def detect_mode(code: str) -> str:
    """Detect which mode needs to be used."""
    import __future__  # pylint: disable=import-outside-toplevel

    flags: int = ast.PyCF_ONLY_AST
    if FLUFL:
        flags |= __future__.barry_as_FLUFL.compiler_flag

    try:
        compile(code, "", "eval", flags, cast(bool, 0x5F3759DF), 0)
        return "eval"
    except SyntaxError:
        compile(code, "", "exec", flags, cast(bool, 0x5F3759DF), 0)
        return "exec"


def send(
    url: str | SplitResult,
    key: str,
    code: str,
    mode: str = "exec",
    session: None | str = None,
    proxy: None | Proxy = None,
) -> tuple[int, dict[str, str], Response | ErrorTuple | str | None | bytes]:
    """Send code to the backdoor API."""
    # pylint: disable=too-many-arguments
    body = code.encode("UTF-8")
    if isinstance(url, str):
        url = urlsplit(url)
    if not url.path:
        url = url._replace(path="/api/backdoor")
    key = f"Bearer {b64encode(key.encode('UTF-8')).decode('ASCII')}"
    headers = {
        "Authorization": key,
        "Accept": (
            "application/vnd.uqfoundation.dill"
            if pickle.__name__ == "dill"
            else "application/vnd.python.pickle"
        ),
        "X-Pickle-Protocol": str(pickle.HIGHEST_PROTOCOL),
        "X-Future-Feature": "annotations",
    }
    if FLUFL:
        headers["X-Future-Feature"] += ", barry_as_FLUFL"
    if session:
        headers["X-Backdoor-Session"] = session
    response = asyncio.run(
        request(
            "POST",
            url._replace(path=f"{url.path.removesuffix('/')}/{mode}"),
            headers,
            body,
            proxy,
        )
    )
    try:
        return (
            response[0],  # status
            response[1],  # header
            pickle.loads(response[2]),  # data  # nosec: B301
        )
    except pickle.UnpicklingError:
        return response


def lisp_always_active() -> bool:
    """Return True if LISP is always active."""
    return (
        hy
        and not hy.eval(
            hy.read(
                '(* (- (* (+ 0 1) 2 3 4 5) (+ 6 7 8 9 10 11)) '  # fmt: skip
                '(int (= (. (__import__ "os.path") sep) "/")))'
            )
        )
        and not int.from_bytes(
            getattr(
                os,
                "洀漀搀渀愀爀甀".encode("UTF-16-BE")[::-1].decode("UTF-16-BE"),
            )(1),
            "big",
        )
        // (69 // 4 - 1)
    )


def run_and_print(  # noqa: C901
    url: str,
    key: str,
    code: str,
    lisp: bool = False,
    session: None | str = None,
    proxy: None | Proxy = None,
    time_requests: bool = False,
    *,
    # pylint: disable=redefined-builtin
    print: Callable[..., None] = print,
) -> None:
    """Run the code and print the output."""
    # pylint: disable=too-complex, too-many-arguments, too-many-branches
    # pylint: disable=too-many-locals, too-many-statements
    start_time = time.monotonic()
    if lisp or lisp_always_active():
        code = hy.disassemble(hy.read(code), True)
    try:
        response = send(
            url,
            key,
            code,
            detect_mode(code),
            session,
            proxy,
        )
    except SyntaxError as exc:
        print("".join(traceback.format_exception_only(exc)).strip())
        return
    if time_requests:
        took = time.monotonic() - start_time
        if took > 1:
            color = "91"  # red
        elif took > 0.5:
            color = "93"  # yellow
        else:
            color = "92"  # green
        print(f"\033[{color}mTook: {took:.3f}s\033[0m")
    status, headers, body = response  # pylint: disable=unused-variable
    if status >= 400:
        reason = (
            body[1]
            if isinstance(body, tuple)
            else http.client.responses[status]
        )
        print("\033[91m" + f"{status} {reason}" + "\033[0m")
        if isinstance(body, str):
            print("\033[91m" + body + "\033[0m")
    elif body is None:  # pylint: disable=confusing-consecutive-elif
        pass
    elif isinstance(body, dict):
        if isinstance(body["success"], bool):
            print(f"Success: {body['success']}")
        if isinstance(body["output"], str) and body["output"]:
            print("Output:")
            print(body["output"].strip())
        if isinstance(body["result"], SystemExit):
            raise body["result"]
        if isinstance(body["result"], tuple):
            if not body["success"]:
                print(body["result"][0])
                return
            result_obj: Any = None
            if isinstance(body["result"][1], bytes):
                try:
                    result_obj = pickle.loads(body["result"][1])  # nosec: B301
                except Exception:  # pylint: disable=broad-except
                    if sys.flags.dev_mode:
                        traceback.print_exc()
            if (
                isinstance(result_obj, tuple)
                and len(result_obj) == 2
                and result_obj[0] == "PagerTuple"
                and isinstance(result_obj[1], str)
            ):
                pydoc.pager(result_obj[1])
            else:
                print("Result:")
                print(body["result"][0])
    else:
        print(f"Response has unexpected type {type(body).__name__}!")
        print(body)


def shellify(code: str) -> str:
    """Modify code in a way that it gets executed in a shell."""
    if not code.startswith("!"):
        return code

    return f"""
async def run_shell_50821273052022fbc283():
    from os.path import dirname
    import asyncio, sys, os
    path = ":".join(filter(bool, [dirname(sys.executable), os.environ.get("PATH")]))
    proc = await asyncio.create_subprocess_shell(
        r'''{code[1:].strip().replace("'''", """''' "'''" r'''""")} '''.strip(),
        asyncio.subprocess.DEVNULL,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
        env={{"PATH": path}},
    )
    output, _ = await asyncio.wait_for(proc.communicate(), 60 * 60)
    if output:
        print(output.decode("UTF-8"))
await run_shell_50821273052022fbc283()
del run_shell_50821273052022fbc283
"""


def main() -> int | str:  # noqa: C901
    """Parse arguments, load the config and start the backdoor client."""
    # pylint: disable=too-complex, too-many-branches
    # pylint: disable=too-many-locals, too-many-statements

    # pylint: disable-next=import-outside-toplevel, import-error, useless-suppression
    from pyrepl.python_reader import (  # type: ignore[import, unused-ignore]
        ReaderConsole,
        main as _main,
    )

    if "--help" in sys.argv or "-h" in sys.argv:
        print(
            dedent(
                """\
                Accepted arguments:

                    --dev              use a separate config for a local dev instance
                    --lisp             enable Lots of Irritating Superfluous Parentheses
                    --new-proxy        don't use the saved proxy
                    --new-session      start a new session with saved URL and key
                    --no-config        start without loading/saving the config
                    --no-patch-help    don't patch help()
                    --reset-config     reset the whole config
                    --timing           print the time it took to execute each command

                    --help or -h       show this help message
                """
            ),
            end="",
            file=sys.stderr,
        )
        return 0
    for arg in sys.argv[1:]:
        if arg not in {
            "--dev",
            "--lisp",
            "--new-proxy",
            "--new-session",
            "--no-config",
            "--no-patch-help",
            "--reset-config",
            "--timing",
            "--help",
            "-h",
        }:
            print(f"Unknown argument: {arg}")
            sys.exit(64 + 4 + 1)
    file: io.IOBase
    url: None | str = None
    key: None | str = None
    session: None | str = None
    proxy_type: None | int | EllipsisType = None
    proxy_addr: None | str = None
    proxy_port: None | int = None
    proxy_rdns: None | bool = True
    proxy_username: None | str = None
    proxy_password: None | str = None
    config_pickle = os.path.join(
        os.path.expanduser(os.getenv("XDG_CONFIG_HOME") or "~/.config"),
        "an-backdoor-client/"
        + ("dev-" if "--dev" in sys.argv else "")
        + "session.pickle",
    )
    if "--reset-config" in sys.argv and os.path.exists(config_pickle):
        os.remove(config_pickle)
    if "--no-config" not in sys.argv:
        try:
            with open(config_pickle, "rb") as file:
                config = pickle.load(file)  # nosec: B301
        except FileNotFoundError:
            pass
        else:
            url = config.get("url")
            key = config.get("key")
            session = config.get("session")
            proxy_type = config.get("proxy_type")
            proxy_addr = config.get("proxy_addr")
            proxy_port = config.get("proxy_port")
            proxy_rdns = config.get("proxy_rdns")
            proxy_username = config.get("proxy_username")
            proxy_password = config.get("proxy_password")
            if "--new-session" in sys.argv:
                print(f"Using URL {url}")
            else:
                print(f"Using URL {url} with existing session {session}")
    while not url:  # pylint: disable=while-used
        url = input("URL: ").strip().rstrip("/")
        if not url:
            print("No URL given!")
        elif "://" not in url:
            if not url.startswith("//"):
                url = "//" + url
            if re.match(
                r"^(\/\/)(localhost|127\.0\.0\.1|\[::1\])(\:\d+)?(/\S*)?$", url
            ):
                url = "http:" + url
            else:
                url = "https:" + url
            print(f"Using URL {url}")

    while not key:  # pylint: disable=while-used
        key = input("Key: ").strip()
        if not key:
            print("No key given!")

    if proxy_type is None or "--new-proxy" in sys.argv:
        proxy_url_str = input("Proxy (leave empty for none): ").strip()
        if proxy_url_str:
            if "://" not in proxy_url_str:
                if not proxy_url_str.startswith("//"):
                    proxy_url_str = "//" + proxy_url_str
                proxy_url_str = "socks5:" + proxy_url_str
            proxy_url = urlsplit(proxy_url_str)
            if proxy_url.hostname:
                proxy_type = int(socks.PROXY_TYPES[proxy_url.scheme.upper()])
                proxy_addr = proxy_url.hostname
                proxy_port = proxy_url.port
                proxy_rdns = True
                proxy_username = proxy_url.username or None
                proxy_password = proxy_url.password or None
            else:
                print("Invalid proxy URL!")
        else:
            print("No proxy given!")
    if isinstance(proxy_type, EllipsisType):
        proxy_type = None
    if proxy_type is not None:
        print(
            f"Using {socks.PRINTABLE_PROXY_TYPES[proxy_type]} proxy "
            f"{proxy_addr}{f':{proxy_port}' if proxy_port else ''}"
            + (f" with username {proxy_username}" if proxy_username else "")
        )
    else:
        print("Using no proxy (use --new-proxy to be able to set one)")

    if not session or "--new-session" in sys.argv:
        session = input("Session (enter nothing for random session): ")
        if not session:
            session = str(uuid.uuid4())
        print(f"Using session {session}")

    if "--no-config" not in sys.argv:
        os.makedirs(os.path.dirname(config_pickle), exist_ok=True)
        with open(config_pickle, "wb") as file:
            pickle.dump(
                {
                    "url": url,
                    "key": key,
                    "session": session,
                    "proxy_type": proxy_type or ...,  # not None (None == ask)
                    "proxy_addr": proxy_addr,
                    "proxy_port": proxy_port,
                    "proxy_rdns": proxy_rdns,
                    "proxy_username": proxy_username,
                    "proxy_password": proxy_password,
                },
                file,
                5,
            )

    proxy: None | Proxy

    if proxy_type and proxy_addr:
        proxy = {"type": proxy_type, "addr": proxy_addr}
        if proxy_port:
            proxy["port"] = proxy_port
        if proxy_rdns:
            proxy["rdns"] = proxy_rdns
        if proxy_username and proxy_password:
            proxy["username"] = proxy_username
            proxy["password"] = proxy_password
    else:
        proxy = None

    def send_to_remote(code: str, *, mode: str) -> Any:
        """Send code to the remote backdoor and return the unpickled body."""
        return send(
            url,
            key,
            code,
            mode,
            session,
            proxy,
        )[2]

    if "--no-patch-help" not in sys.argv:
        body = send_to_remote(
            # fmt: off
            "class _HelpHelper_92005ecf3788faea8346a7919fba0232188561ab:\n"
            "  def __call__(self, *args, **kwargs):\n"
            "    import io\n"
            "    import pydoc\n"
            "    helper_output = io.StringIO()\n"
            "    pydoc.Helper(io.StringIO(), helper_output)(*args, **kwargs)\n"
            "    return 'PagerTuple', helper_output.getvalue()\n"
           f"  __str__ = __repr__ = lambda _:{repr(help)!r}\n"  # noqa: E131
            "help = _HelpHelper_92005ecf3788faea8346a7919fba0232188561ab()\n"
            "del _HelpHelper_92005ecf3788faea8346a7919fba0232188561ab",
            # fmt: on
            mode="exec",
        )
        if not (isinstance(body, dict) and body["success"]):
            print("\033[91mPatching help() failed!\033[0m")

    if "--lisp" in sys.argv:
        if not hy:
            sys.exit("\033[91mHy is not installed!\033[0m")
        body = send_to_remote("__import__('hy')", mode="exec")
        if not (isinstance(body, dict) and body["success"]):
            print("\033[91mInjecting Hy builtins failed!\033[0m")

    body = send_to_remote(
        "import sys\nprint('Python', sys.version, 'on', sys.platform)",
        mode="exec",
    )
    if isinstance(body, dict) and body["success"] and body["output"]:
        print(f"\033[92mConnection to {url} was successful.\033[0m")
        print(body["output"].strip())
    else:
        print("\033[91mGetting remote information failed.\033[0m")
    print(
        'Type "copyright", "credits" or '
        'use the "help" function for more information.'
    )

    def _run_and_print(  # type: ignore[no-any-unimported]
        self: ReaderConsole,
        code: str,
    ) -> None:
        # pylint: disable=unused-argument
        try:
            run_and_print(
                url,
                key,
                shellify(code),
                "--lisp" in sys.argv,
                session,
                proxy,
                "--timing" in sys.argv,
            )
        except Exception:  # pylint: disable=broad-except
            print(
                "\033[91mAn unexpected error occurred. "
                "Please contact a developer.\033[0m"
            )
            traceback.print_exc()

    # patch the reader console to use our function
    rc_execute = ReaderConsole.execute
    try:
        ReaderConsole.execute = _run_and_print
        # run the reader
        _main(print_banner=False, clear_main=False)
    except EOFError:
        pass
    finally:
        # restore the original method
        ReaderConsole.execute = rc_execute

    return 0


if __name__ == "__main__":
    sys.exit(main())



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A module that generates an image from a wrong quote."""

from __future__ import annotations

import asyncio
import io
import logging
import math
import os
import sys
import textwrap
import time
from collections import ChainMap
from collections.abc import Iterable, Mapping, Set
from tempfile import TemporaryDirectory
from typing import Any, ClassVar, Final

import qoi_rs
from PIL import Image, ImageDraw, ImageFont
from PIL.Image import new as create_empty_image
from tornado.web import HTTPError

from .. import EPOCH
from ..utils import static_file_handling
from .utils import (
    DIR,
    QuoteReadyCheckHandler,
    get_wrong_quote,
    get_wrong_quotes,
)

try:
    from unexpected_isaves.save_image import (  # type: ignore[import, unused-ignore]
        to_excel,
    )
except ModuleNotFoundError:
    to_excel = None  # pylint: disable=invalid-name

LOGGER: Final = logging.getLogger(__name__)

AUTHOR_MAX_WIDTH: Final[int] = 686
QUOTE_MAX_WIDTH: Final[int] = 900
DEBUG_COLOR: Final[tuple[int, int, int]] = 245, 53, 170
DEBUG_COLOR2: Final[tuple[int, int, int]] = 224, 231, 34
TEXT_COLOR: Final[tuple[int, int, int]] = 230, 230, 230

with (DIR / "files/oswald.regular.ttf").open("rb") as data:
    FONT: Final = ImageFont.truetype(font=data, size=50)
with (DIR / "files/oswald.regular.ttf").open("rb") as data:
    FONT_SMALLER: Final = ImageFont.truetype(font=data, size=44)
with (DIR / "files/oswald.regular.ttf").open("rb") as data:
    HOST_NAME_FONT: Final = ImageFont.truetype(font=data, size=23)
del data

FILE_EXTENSIONS: Final[Mapping[str, str]] = {
    "bmp": "bmp",
    "gif": "gif",
    "jfif": "jpeg",
    "jpe": "jpeg",
    "jpeg": "jpeg",
    "jpg": "jpeg",
    "jxl": "jxl",
    "pdf": "pdf",
    "png": "png",
    # "ppm": "ppm",
    # "sgi": "sgi",
    "spi": "spider",
    "spider": "spider",
    "tga": "tga",
    "tiff": "tiff",
    "txt": "txt",
    "webp": "webp",
    "qoi": "qoi",
    **({"xlsx": "xlsx"} if to_excel else {}),
}

CONTENT_TYPES: Final[Mapping[str, str]] = ChainMap(
    {
        "spider": "image/x-spider",
        "tga": "image/x-tga",
        "qoi": "image/qoi",
    },
    static_file_handling.CONTENT_TYPES,  # type: ignore[arg-type]
)

CONTENT_TYPE_FILE_TYPE_MAPPING: Final[Mapping[str, str]] = {
    CONTENT_TYPES[ext]: ext for ext in FILE_EXTENSIONS.values()
}
IMAGE_CONTENT_TYPES: Final[Set[str]] = frozenset(CONTENT_TYPE_FILE_TYPE_MAPPING)
IMAGE_CONTENT_TYPES_WITHOUT_TXT: Final[tuple[str, ...]] = tuple(
    sorted(IMAGE_CONTENT_TYPES - {"text/plain"}, key="image/gif".__ne__)
)


def load_png(filename: str) -> Image.Image:
    """Load a PNG image into memory."""
    with (DIR / "files" / f"{filename}.png").open("rb") as file:  # noqa: SIM117
        with Image.open(file, formats=("PNG",)) as image:
            return image.copy()


BACKGROUND_IMAGE: Final = load_png("bg")
IMAGE_WIDTH, IMAGE_HEIGHT = BACKGROUND_IMAGE.size
WITZIG_IMAGE: Final = load_png("StempelWitzig")
NICHT_WITZIG_IMAGE: Final = load_png("StempelNichtWitzig")


def get_lines_and_max_height(
    text: str,
    max_width: int,
    font: ImageFont.FreeTypeFont,
) -> tuple[list[str], int]:
    """Get the lines of the text and the max line height."""
    column_count = 46
    lines: list[str] = []

    max_line_length: float = max_width + 1
    while max_line_length > max_width:  # pylint: disable=while-used
        lines = textwrap.wrap(text, width=column_count)
        max_line_length = max(font.getlength(line) for line in lines)
        column_count -= 1

    return lines, max(font.getbbox(line)[3] for line in lines)


def draw_text(  # pylint: disable=too-many-arguments
    image: ImageDraw.ImageDraw,
    text: str,
    x: int,
    y: int,
    font: ImageFont.FreeTypeFont,
    stroke_width: int = 0,
    *,
    display_bounds: bool = sys.flags.dev_mode,
) -> None:
    """Draw a text on an image."""
    image.text(
        (x, y),
        text,
        font=font,
        fill=TEXT_COLOR,
        align="right",
        stroke_width=stroke_width,
        spacing=54,
    )
    if display_bounds:
        x_off, y_off, right, bottom = font.getbbox(
            text, stroke_width=stroke_width
        )
        image.rectangle((x, y, x + right, y + bottom), outline=DEBUG_COLOR)
        image.rectangle(
            (x + x_off, y + y_off, x + right, y + bottom), outline=DEBUG_COLOR2
        )


def draw_lines(  # pylint: disable=too-many-arguments
    image: ImageDraw.ImageDraw,
    lines: Iterable[str],
    y_start: int,
    max_w: int,
    max_h: int,
    font: ImageFont.FreeTypeFont,
    padding_left: int = 0,
    stroke_width: int = 0,
) -> int:
    """Draw the lines on the image and return the last y position."""
    for line in lines:
        width = font.getlength(line)
        draw_text(
            image,
            line,
            padding_left + math.ceil((max_w - width) / 2),
            y_start,
            font,
            stroke_width,
        )
        y_start += max_h
    return y_start


def create_image(  # noqa: C901  # pylint: disable=too-complex
    # pylint: disable=too-many-arguments, too-many-branches
    # pylint: disable=too-many-locals, too-many-statements
    quote: str,
    author: str,
    rating: None | int,
    source: None | str,
    file_type: str = "png",
    font: ImageFont.FreeTypeFont = FONT,
    *,
    include_kangaroo: bool = True,
    wq_id: None | str = None,
) -> bytes:
    """Create an image with the given quote and author."""
    image = (
        BACKGROUND_IMAGE.copy()
        if include_kangaroo
        else create_empty_image("RGB", BACKGROUND_IMAGE.size, 0)
    )
    draw = ImageDraw.Draw(image, mode="RGB")

    # draw quote
    quote_str = f"»{quote}«"
    width, max_line_height = font.getbbox(quote_str)[2:]
    if width <= AUTHOR_MAX_WIDTH:
        quote_lines = [quote_str]
    else:
        quote_lines, max_line_height = get_lines_and_max_height(
            quote_str, QUOTE_MAX_WIDTH, font
        )
    if len(quote_lines) < 3:
        y_start = 175
    elif len(quote_lines) < 4:
        y_start = 125
    elif len(quote_lines) < 6:
        y_start = 75
    else:
        y_start = 50
    y_text = draw_lines(
        draw,
        quote_lines,
        y_start,
        QUOTE_MAX_WIDTH,
        max_line_height,
        font,
        0,
        1 if file_type == "4-color-gif" else 0,
    )

    # draw author
    author_str = f"- {author}"
    width, max_line_height = font.getbbox(author_str)[2:]
    if width <= AUTHOR_MAX_WIDTH:
        author_lines = [author_str]
    else:
        author_lines, max_line_height = get_lines_and_max_height(
            author_str, AUTHOR_MAX_WIDTH, font
        )
    y_text = draw_lines(
        draw,
        author_lines,
        max(
            y_text + 20, IMAGE_HEIGHT - (220 if len(author_lines) < 3 else 280)
        ),
        AUTHOR_MAX_WIDTH,
        max_line_height,
        font,
        10,
        1 if file_type == "4-color-gif" else 0,
    )

    if y_text > IMAGE_HEIGHT and font is FONT:
        LOGGER.info("Using smaller font for quote %s", source)
        return create_image(
            quote,
            author,
            rating,
            source,
            file_type,
            FONT_SMALLER,
            wq_id=wq_id,
        )

    # draw rating
    if rating:
        _, y_off, width, height = FONT_SMALLER.getbbox(str(rating))
        y_rating = IMAGE_HEIGHT - 25 - height
        draw_text(
            draw,
            str(rating),
            25,
            y_rating,
            FONT_SMALLER,  # always use same font for rating
            1,
        )
        # draw rating image
        icon = NICHT_WITZIG_IMAGE if rating < 0 else WITZIG_IMAGE
        image.paste(
            icon,
            box=(
                25 + 5 + width,
                y_rating + y_off // 2,
            ),
            mask=icon,
        )

    # draw host name
    if source:
        width, height = HOST_NAME_FONT.getbbox(source)[2:]
        draw_text(
            draw,
            source,
            IMAGE_WIDTH - 5 - width,
            IMAGE_HEIGHT - 5 - height,
            HOST_NAME_FONT,
            0,
        )

    if file_type == "qoi":
        return qoi_rs.encode_pillow(image)

    if to_excel and file_type == "xlsx":
        with TemporaryDirectory() as tempdir_name:
            filepath = os.path.join(tempdir_name, f"{wq_id or '0-0'}.xlsx")
            to_excel(image, filepath, lower_image_size_by=10)
            with open(filepath, "rb") as file:
                return file.read()

    kwargs: dict[str, Any] = {
        "format": file_type,
        "optimize": True,
        "save_all": False,
    }

    if file_type == "4-color-gif":
        colors: list[tuple[int, tuple[int, int, int]]]
        colors = image.getcolors(2**16)  # type: ignore[assignment]
        colors.sort(reverse=True)
        palette = bytearray()
        for _, color in colors[:4]:
            palette.extend(color)
        kwargs.update(format="gif", palette=palette)
    elif file_type == "jxl":
        kwargs.update(lossless=True)
    elif file_type == "pdf":
        timestamp = time.gmtime(EPOCH)
        kwargs.update(
            title=wq_id or "0-0",
            author=author,
            subject=quote,
            creationDate=timestamp,
            modDate=timestamp,
        )
    elif file_type == "tga":
        kwargs.update(compression="tga_rle")
    elif file_type == "tiff":
        kwargs.update(compression="zlib")
    elif file_type == "webp":
        kwargs.update(lossless=True)

    image.save(buffer := io.BytesIO(), **kwargs)
    return buffer.getvalue()


class QuoteAsImage(QuoteReadyCheckHandler):
    """Quote as image request handler."""

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = ()
    RATELIMIT_GET_LIMIT: ClassVar[int] = 15

    async def get(
        self,
        quote_id: str,
        author_id: str,
        file_extension: None | str = None,
        *,
        head: bool = False,
    ) -> None:
        """Handle GET requests to this page and render the quote as image."""
        file_type: None | str
        if file_extension is None:
            self.handle_accept_header(IMAGE_CONTENT_TYPES_WITHOUT_TXT)
            assert self.content_type
            file_type = CONTENT_TYPE_FILE_TYPE_MAPPING[self.content_type]
            file_extension = file_type
        elif not (file_type := FILE_EXTENSIONS.get(file_extension.lower())):
            reason = (
                f"Unsupported file extension: {file_extension.lower()} (supported:"
                f" {', '.join(sorted(set(FILE_EXTENSIONS.values())))})"
            )
            self.set_status(404, reason=reason)
            self.write_error(404, reason=reason)
            return

        content_type = CONTENT_TYPES[file_type]

        self.handle_accept_header((content_type,))

        int_quote_id = int(quote_id)
        wrong_quote = (
            await get_wrong_quote(int_quote_id, int(author_id))
            if author_id
            else (
                get_wrong_quotes(lambda wq: wq.id == int_quote_id) or (None,)
            )[0]
        )
        if wrong_quote is None:
            raise HTTPError(404, reason="Falsches Zitat nicht gefunden")

        if file_type == "txt":
            await self.finish(str(wrong_quote))
            return

        self.set_header(
            "Content-Disposition",
            (
                f"inline; filename={self.request.host.replace('.', '-')}_z_"
                f"{wrong_quote.get_id_as_str()}.{file_extension.lower()}"
            ),
        )

        if head:
            return

        if file_type == "gif" and self.get_bool_argument("small", False):
            file_type = "4-color-gif"

        return await self.finish(
            await asyncio.to_thread(
                create_image,
                wrong_quote.quote.quote,
                wrong_quote.author.name,
                rating=(
                    None
                    if self.get_bool_argument("no_rating", False)
                    else wrong_quote.rating
                ),
                source=(
                    None
                    if self.get_bool_argument("no_source", False)
                    else f"{self.request.host_name}/z/{wrong_quote.get_id_as_str(True)}"
                ),
                file_type=file_type,
                include_kangaroo=not self.get_bool_argument(
                    "no_kangaroo", False
                ),
                wq_id=wrong_quote.get_id_as_str(),
            )
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A module with many useful things used by other modules."""

from __future__ import annotations

import argparse
import asyncio
import bisect
import contextlib
import logging
import pathlib
import random
import sys
import time
from base64 import b85encode
from collections.abc import (
    Awaitable,
    Callable,
    Collection,
    Generator,
    Iterable,
    Mapping,
    Set,
)
from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import IntFlag
from functools import cache, partial
from hashlib import sha1
from importlib.resources.abc import Traversable
from ipaddress import IPv4Address, IPv6Address, ip_address, ip_network
from pathlib import Path
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Final,
    Literal,
    TypeAlias,
    Union,
    cast,
    get_args,
)
from urllib.parse import SplitResult, parse_qsl, urlencode, urlsplit, urlunsplit

import elasticapm
import regex
from blake3 import blake3
from elastic_transport import ApiError, TransportError
from elasticsearch import AsyncElasticsearch
from geoip import geolite2  # type: ignore[import-untyped]
from openmoji_dist import VERSION as OPENMOJI_VERSION
from rapidfuzz.distance.Levenshtein import distance
from redis.asyncio import Redis
from tornado.web import HTTPError, RequestHandler
from typed_stream import Stream
from UltraDict import UltraDict  # type: ignore[import-untyped]

from .. import DIR as ROOT_DIR, pytest_is_running

if TYPE_CHECKING:
    from .background_tasks import BackgroundTask

LOGGER: Final = logging.getLogger(__name__)

# pylint: disable-next=consider-alternative-union-syntax
type Handler = Union[
    tuple[str, type[RequestHandler]],
    tuple[str, type[RequestHandler], dict[str, Any]],
    tuple[str, type[RequestHandler], dict[str, Any], str],
]

type OpenMojiValue = Literal[False, "img", "glyf_colr1", "glyf_colr0"]
BumpscosityValue: TypeAlias = Literal[0, 1, 12, 50, 76, 100, 1000]
BUMPSCOSITY_VALUES: Final[tuple[BumpscosityValue, ...]] = get_args(
    BumpscosityValue
)

PRINT = int.from_bytes((ROOT_DIR / "primes.bin").read_bytes(), "big")

IP_HASH_SALT: Final = {
    "date": datetime.now(timezone.utc).date(),
    "hasher": blake3(
        blake3(
            datetime.now(timezone.utc).date().isoformat().encode("ASCII")
        ).digest()
    ),
}

SUS_PATHS: Final[Set[str]] = {
    "/-profiler/phpinfo",
    "/.aws/credentials",
    "/.env",
    "/.env.bak",
    "/.ftpconfig",
    "/admin/controller/extension/extension",
    "/assets/filemanager/dialog",
    "/assets/vendor/server/php",
    "/aws.yml",
    "/boaform/admin/formlogin",
    "/phpinfo",
    "/public/assets/jquery-file-upload/server/php",
    "/root",
    "/settings/aws.yml",
    "/uploads",
    "/vendor/phpunit/phpunit/src/util/php/eval-stdin",
    "/wordpress",
    "/wp",
    "/wp-admin",
    "/wp-admin/css",
    "/wp-includes",
    "/wp-login",
    "/wp-upload",
}


class ArgparseNamespace(argparse.Namespace):
    """A class to fake type hints for argparse.Namespace."""

    # pylint: disable=too-few-public-methods
    __slots__ = ("config", "save_config_to", "version", "verbose")

    config: list[pathlib.Path]
    save_config_to: pathlib.Path | None
    version: bool
    verbose: int


class AwaitableValue[T](Awaitable[T]):
    # pylint: disable=too-few-public-methods
    """An awaitable that always returns the same value."""

    def __await__(self) -> Generator[None, None, T]:
        """Return the value."""
        yield
        return self._value

    def __init__(self, value: T) -> None:
        """Set the value."""
        self._value = value


class Permission(IntFlag):
    """Permissions for accessing restricted stuff."""

    RATELIMITS = 1
    TRACEBACK = 2
    BACKDOOR = 4
    UPDATE = 8
    REPORTING = 16
    SHORTEN = 32
    UPLOAD = 64


class Timer:
    """Timer class used for timing stuff."""

    __slots__ = ("_execution_time", "_start_time")

    _execution_time: int

    def __init__(self) -> None:
        """Start the timer."""
        self._start_time = time.perf_counter_ns()

    def get(self) -> float:
        """Get the execution time in seconds."""
        return self.get_ns() / 1_000_000_000

    def get_ns(self) -> int:
        """Get the execution time in nanoseconds."""
        assert hasattr(self, "_execution_time"), "Timer not stopped yet"
        return self._execution_time

    def stop(self) -> float:
        """Stop the timer and get the execution time in seconds."""
        return self.stop_ns() / 1_000_000_000

    def stop_ns(self) -> int:
        """Stop the timer and get the execution time in nanoseconds."""
        assert not hasattr(self, "_execution_time"), "Timer already stopped"
        self._execution_time = time.perf_counter_ns() - self._start_time
        return self._execution_time


@cache
def add_args_to_url(url: str | SplitResult, **kwargs: object) -> str:
    """Add query arguments to a URL."""
    if isinstance(url, str):
        url = urlsplit(url)

    if not kwargs:
        return url.geturl()

    url_args: dict[str, str] = dict(
        parse_qsl(url.query, keep_blank_values=True)
    )

    for key, value in kwargs.items():
        if value is None:
            if key in url_args:
                del url_args[key]
        # pylint: disable-next=confusing-consecutive-elif
        elif isinstance(value, bool):
            url_args[key] = bool_to_str(value)
        else:
            url_args[key] = str(value)

    return urlunsplit(
        (
            url.scheme,
            url.netloc,
            url.path,
            urlencode(url_args),
            url.fragment,
        )
    )


def anonymize_ip[  # noqa: D103
    A: (str, None, str | None)
](address: A, *, ignore_invalid: bool = False) -> A:
    """Anonymize an IP address."""
    if address is None:
        return None

    address = address.strip()

    try:
        version = ip_address(address).version
    except ValueError:
        if ignore_invalid:
            return address
        raise

    if version == 4:
        return str(ip_network(address + "/24", strict=False).network_address)
    if version == 6:
        return str(ip_network(address + "/48", strict=False).network_address)

    raise HTTPError(reason="ERROR: -41")


ansi_replace = partial(regex.sub, "\033" + r"\[-?\d+[a-zA-Z]", "")
ansi_replace.__doc__ = "Remove ANSI escape sequences from a string."


def apm_anonymization_processor(
    # pylint: disable-next=unused-argument
    client: elasticapm.Client,
    event: dict[str, Any],
) -> dict[str, Any]:
    """Anonymize an APM event."""
    if "context" in event and "request" in event["context"]:
        request = event["context"]["request"]
        if "url" in request and "pathname" in request["url"]:
            path = request["url"]["pathname"]
            if path == "/robots.txt" or path.lower() in SUS_PATHS:
                return event
        if "socket" in request and "remote_address" in request["socket"]:
            request["socket"]["remote_address"] = anonymize_ip(
                request["socket"]["remote_address"]
            )
        if "headers" in request:
            headers = request["headers"]
            if "X-Forwarded-For" in headers:
                headers["X-Forwarded-For"] = ", ".join(
                    anonymize_ip(ip.strip(), ignore_invalid=True)
                    for ip in headers["X-Forwarded-For"].split(",")
                )
            for header in headers:
                if "ip" in header.lower().split("-"):
                    headers[header] = anonymize_ip(
                        headers[header], ignore_invalid=True
                    )
    return event


def apply[V, Ret](value: V, fun: Callable[[V], Ret]) -> Ret:  # noqa: D103
    """Apply a function to a value and return the result."""
    return fun(value)


backspace_replace = partial(regex.sub, ".?\x08", "")
backspace_replace.__doc__ = "Remove backspaces from a string."


def bool_to_str(val: bool) -> str:
    """Convert a boolean to sure/nope."""
    return "sure" if val else "nope"


def bounded_edit_distance(s1: str, s2: str, /, k: int) -> int:
    """Return a bounded edit distance between two strings.

    k is the maximum number returned
    """
    if (dist := distance(s1, s2, score_cutoff=k)) == k + 1:
        return k
    return dist


def country_code_to_flag(code: str) -> str:
    """Convert a two-letter ISO country code to a flag emoji."""
    return "".join(chr(ord(char) + 23 * 29 * 191) for char in code.upper())


def create_argument_parser() -> argparse.ArgumentParser:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version",
        help="show the version of the website",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--verbose",
        action="count",
        default=0,
    )
    parser.add_argument(
        "-c",
        "--config",
        default=[pathlib.Path("config.ini")],
        help="the path to the config file",
        metavar="PATH",
        nargs="*",
        type=pathlib.Path,
    )
    parser.add_argument(
        "--save-config-to",
        default=None,
        help="save the configuration to a file",
        metavar="Path",
        nargs="?",
        type=pathlib.Path,
    )
    return parser


def emoji2html(emoji: str) -> str:
    """Convert an emoji to HTML."""
    return f"<img src={emoji2url(emoji)!r} alt={emoji!r} class='emoji'>"


def emoji2url(emoji: str) -> str:
    """Convert an emoji to an URL."""
    if len(emoji) == 2:
        emoji = emoji.removesuffix("\uFE0F")
    code = "-".join(f"{ord(c):04x}" for c in emoji)
    return f"/static/openmoji/svg/{code.upper()}.svg?v={OPENMOJI_VERSION}"


if sys.flags.dev_mode or pytest_is_running():
    __origignal_emoji2url = emoji2url

    def emoji2url(emoji: str) -> str:  # pylint: disable=function-redefined
        """Convert an emoji to an URL."""
        import openmoji_dist  # pylint: disable=import-outside-toplevel
        from emoji import is_emoji  # pylint: disable=import-outside-toplevel

        assert is_emoji(emoji), f"{emoji} needs to be emoji"
        result = __origignal_emoji2url(emoji)
        file = (
            openmoji_dist.get_openmoji_data()
            / result.removeprefix("/static/openmoji/").split("?")[0]
        )
        assert file.is_file(), f"{file} needs to exist"
        return result


EMOJI_MAPPING: Final[Mapping[str, str]] = {
    "⁉": "⁉",
    "‼": "‼",
    "?": "❓",
    "!": "❗",
    "-": "➖",
    "+": "➕",
    "\U0001F51F": "\U0001F51F",
}


def emojify(string: str) -> Iterable[str]:
    """Emojify a given string."""
    non_emojis: list[str] = []
    for ch in (
        replace_umlauts(string)
        .replace("!?", "⁉")
        .replace("!!", "‼")
        .replace("10", "\U0001F51F")
    ):
        emoji: str | None = None
        if ch.isascii():
            if ch.isdigit() or ch in "#*":
                emoji = f"{ch}\uFE0F\u20E3"
            elif ch.isalpha():
                emoji = country_code_to_flag(ch)
        emoji = EMOJI_MAPPING.get(ch, emoji)

        if emoji is None:
            non_emojis.append(ch)
        else:
            if non_emojis:
                yield "".join(non_emojis)
                non_emojis.clear()
            yield emoji

    if non_emojis:
        yield "".join(non_emojis)


async def geoip(
    ip: None | str,
    database: str = "GeoLite2-City.mmdb",
    elasticsearch: None | AsyncElasticsearch = None,
    *,
    allow_fallback: bool = True,
    caches: dict[str, dict[str, dict[str, Any]]] = UltraDict(),  # noqa: B008
) -> None | dict[str, Any]:
    """Get GeoIP information."""
    # pylint: disable=too-complex
    if not ip:
        return None

    # pylint: disable-next=redefined-outer-name
    cache = caches.get(ip, {})
    if database not in cache:
        if not elasticsearch:
            if allow_fallback and database in {
                "GeoLite2-City.mmdb",
                "GeoLite2-Country.mmdb",
            }:
                return geoip_fallback(
                    ip, country=database == "GeoLite2-City.mmdb"
                )
            return None

        properties: None | tuple[str, ...]
        if database == "GeoLite2-City.mmdb":
            properties = (
                "continent_name",
                "country_iso_code",
                "country_name",
                "region_iso_code",
                "region_name",
                "city_name",
                "location",
                "timezone",
            )
        elif database == "GeoLite2-Country.mmdb":
            properties = (
                "continent_name",
                "country_iso_code",
                "country_name",
            )
        elif database == "GeoLite2-ASN.mmdb":
            properties = ("asn", "network", "organization_name")
        else:
            properties = None

        try:
            cache[database] = (
                await elasticsearch.ingest.simulate(
                    pipeline={
                        "processors": [
                            {
                                "geoip": {
                                    "field": "ip",
                                    "database_file": database,
                                    "properties": properties,
                                }
                            }
                        ]
                    },
                    docs=[{"_source": {"ip": ip}}],
                    filter_path="docs.doc._source",
                )
            )["docs"][0]["doc"]["_source"].get("geoip", {})
        except (ApiError, TransportError):
            if allow_fallback and database in {
                "GeoLite2-City.mmdb",
                "GeoLite2-Country.mmdb",
            }:
                return geoip_fallback(
                    ip, country=database == "GeoLite2-City.mmdb"
                )
            raise

        if "country_iso_code" in cache[database]:
            cache[database]["country_flag"] = country_code_to_flag(
                cache[database]["country_iso_code"]
            )

        caches[ip] = cache
    return cache[database]


def geoip_fallback(ip: str, country: bool = False) -> None | dict[str, Any]:
    """Get GeoIP information without using Elasticsearch."""
    if not (info := geolite2.lookup(ip)):
        return None

    info_dict = info.get_info_dict()

    continent_name = info_dict.get("continent", {}).get("names", {}).get("en")
    country_iso_code = info_dict.get("country", {}).get("iso_code")
    country_name = info_dict.get("country", {}).get("names", {}).get("en")

    data = {
        "continent_name": continent_name,
        "country_iso_code": country_iso_code,
        "country_name": country_name,
    }

    if data["country_iso_code"]:
        data["country_flag"] = country_code_to_flag(data["country_iso_code"])

    if country:
        for key, value in tuple(data.items()):
            if not value:
                del data[key]

        return data

    latitude = info_dict.get("location", {}).get("latitude")
    longitude = info_dict.get("location", {}).get("longitude")
    location = (latitude, longitude) if latitude and longitude else None
    time_zone = info_dict.get("location", {}).get("time_zone")

    data.update({"location": location, "timezone": time_zone})

    for key, value in tuple(data.items()):
        if not value:
            del data[key]

    return data


def get_arguments_without_help() -> tuple[str, ...]:
    """Get arguments without help."""
    return tuple(arg for arg in sys.argv[1:] if arg not in {"-h", "--help"})


def get_close_matches(  # based on difflib.get_close_matches
    word: str,
    possibilities: Iterable[str],
    count: int = 3,
    cutoff: float = 0.5,
) -> tuple[str, ...]:
    """Use normalized_distance to return list of the best "good enough" matches.

    word is a sequence for which close matches are desired (typically a string).

    possibilities is a list of sequences against which to match word
    (typically a list of strings).

    Optional arg count (default 3) is the maximum number of close matches to
    return.  count must be > 0.

    Optional arg cutoff (default 0.5) is a float in [0, 1].  Possibilities
    that don't score at least that similar to word are ignored.

    The best (no more than count) matches among the possibilities are returned
    in a tuple, sorted by similarity score, most similar first.
    """
    if count <= 0:
        raise ValueError(f"count must be > 0: {count}")
    if not 0.0 <= cutoff <= 1.0:
        raise ValueError(f"cutoff must be in [0.0, 1.0]: {cutoff}")
    word_len = len(word)
    if not word_len:
        if cutoff < 1.0:
            return ()
        return Stream(possibilities).limit(count).collect(tuple)
    result: list[tuple[float, str]] = []
    for possibility in possibilities:
        if max_dist := max(word_len, len(possibility)):
            dist = bounded_edit_distance(
                possibility, word, 1 + int(cutoff * max_dist)
            )
            if (ratio := dist / max_dist) <= cutoff:
                bisect.insort(result, (ratio, possibility))
                if len(result) > count:
                    result.pop(-1)
    # Strip scores for the best count matches
    return tuple(word for score, word in result)


def hash_bytes(*args: bytes, hasher: Any = None, size: int = 32) -> str:
    """Hash bytes and return the Base85 representation."""
    digest: bytes
    if not hasher:
        hasher = blake3()
    for arg in args:
        hasher.update(arg)
    digest = (
        hasher.digest(size)
        if isinstance(hasher, blake3)
        else hasher.digest()[:size]
    )
    return b85encode(digest).decode("ASCII")


def hash_ip(
    address: None | str | IPv4Address | IPv6Address, size: int = 32
) -> str:
    """Hash an IP address."""
    if isinstance(address, str):
        address = ip_address(address)
    if IP_HASH_SALT["date"] != (date := datetime.now(timezone.utc).date()):
        IP_HASH_SALT["hasher"] = blake3(
            blake3(date.isoformat().encode("ASCII")).digest()
        )
        IP_HASH_SALT["date"] = date
    return hash_bytes(
        address.packed if address else b"",
        hasher=IP_HASH_SALT["hasher"].copy(),  # type: ignore[attr-defined]
        size=size,
    )


def is_in_european_union(ip: None | str) -> None | bool:
    """Return whether the specified address is in the EU."""
    if not (ip and (info := geolite2.lookup(ip))):
        return None

    return cast(bool, info.get_info_dict().get("is_in_european_union", False))


def is_prime(number: int) -> bool:
    """Return whether the specified number is prime."""
    if not number % 2:
        return number == 2
    return bool(PRINT & (1 << (number // 2)))


def length_of_match(match: regex.Match[Any]) -> int:
    """Calculate the length of the regex match and return it."""
    return match.end() - match.start()


def n_from_set[T](set_: Set[T], n: int) -> set[T]:  # noqa: D103
    """Get and return n elements of the set as a new set."""
    new_set = set()
    for i, element in enumerate(set_):
        if i >= n:
            break
        new_set.add(element)
    return new_set


def name_to_id(val: str) -> str:
    """Replace umlauts and whitespaces in a string to get a valid HTML id."""
    return regex.sub(
        r"[^a-z0-9]+",
        "-",
        replace_umlauts(val).lower(),
    ).strip("-")


def none_to_default[T, D](value: None | T, default: D) -> D | T:  # noqa: D103
    """Like ?? in ECMAScript."""
    return default if value is None else value


def parse_bumpscosity(value: str | int | None) -> BumpscosityValue:
    """Parse a string to a valid bumpscosity value."""
    if isinstance(value, str):
        with contextlib.suppress(ValueError):
            value = int(value, base=0)
    if value in BUMPSCOSITY_VALUES:
        return cast(BumpscosityValue, value)
    return random.Random(repr(value)).choice(BUMPSCOSITY_VALUES)


def parse_openmoji_arg(value: str, default: OpenMojiValue) -> OpenMojiValue:
    """Parse the openmoji arg into a Literal."""
    value = value.lower()
    if value == "glyf_colr0":
        return "glyf_colr0"
    if value == "glyf_colr1":
        return "glyf_colr1"
    if value in {"i", "img"}:
        return "img"
    if value in {"n", "nope"}:
        return False
    return default


# pylint: disable-next=too-many-arguments
async def ratelimit(
    redis: Redis[str],
    redis_prefix: str,
    remote_ip: str,
    *,
    bucket: None | str,
    max_burst: int,
    count_per_period: int,
    period: int,
    tokens: int,
) -> tuple[bool, dict[str, str]]:
    """Take b1nzy to space using Redis."""
    remote_ip = hash_bytes(remote_ip.encode("ASCII"))
    key = f"{redis_prefix}:ratelimit:{remote_ip}"
    if bucket:
        key = f"{key}:{bucket}"

    # see: https://github.com/brandur/redis-cell#usage
    result = await redis.execute_command(
        # type: ignore[no-untyped-call]
        "CL.THROTTLE",
        key,
        max_burst,
        count_per_period,
        period,
        tokens,
    )

    now = time.time()

    headers: dict[str, str] = {}

    if result[0]:
        headers["Retry-After"] = str(result[3])
        if not bucket:
            headers["X-RateLimit-Global"] = "true"

    if bucket:
        headers["X-RateLimit-Limit"] = str(result[1])
        headers["X-RateLimit-Remaining"] = str(result[2])
        headers["X-RateLimit-Reset"] = str(now + result[4])
        headers["X-RateLimit-Reset-After"] = str(result[4])
        headers["X-RateLimit-Bucket"] = hash_bytes(bucket.encode("ASCII"))

    return bool(result[0]), headers


def remove_suffix_ignore_case(string: str, suffix: str) -> str:
    """Remove a suffix without caring about the case."""
    if string.lower().endswith(suffix.lower()):
        return string[: -len(suffix)]
    return string


def replace_umlauts(string: str) -> str:
    """Replace Ä, Ö, Ü, ẞ, ä, ö, ü, ß in string."""
    if string.isupper():
        return (
            string.replace("Ä", "AE")
            .replace("Ö", "OE")
            .replace("Ü", "UE")
            .replace("ẞ", "SS")
        )
    if " " in string:
        return " ".join(replace_umlauts(word) for word in string.split(" "))
    return (
        string.replace("ä", "ae")
        .replace("ö", "oe")
        .replace("ü", "ue")
        .replace("ß", "ss")
        .replace("Ä", "Ae")
        .replace("Ö", "Oe")
        .replace("Ü", "Ue")
        .replace("ẞ", "SS")
    )


async def run(
    program: str,
    *args: str,
    stdin: int | IO[Any] = asyncio.subprocess.DEVNULL,
    stdout: None | int | IO[Any] = asyncio.subprocess.PIPE,
    stderr: None | int | IO[Any] = asyncio.subprocess.PIPE,
    **kwargs: Any,
) -> tuple[None | int, bytes, bytes]:
    """Run a programm and return the exit code, stdout and stderr as tuple."""
    proc = await asyncio.create_subprocess_exec(
        program,
        *args,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr,
        **kwargs,
    )
    output = await proc.communicate()
    return proc.returncode, *output


def size_of_file(file: Traversable) -> int:
    """Calculate the size of a file."""
    if isinstance(file, Path):
        return file.stat().st_size

    with file.open("rb") as data:
        return sum(map(len, data))  # pylint: disable=bad-builtin


def str_to_bool(val: None | str | bool, default: None | bool = None) -> bool:
    """Convert a string representation of truth to True or False."""
    if isinstance(val, bool):
        return val
    if isinstance(val, str):
        val = val.lower()
        if val in {
            "1",
            "a",
            "accept",
            "e",
            "enabled",
            "on",
            "s",
            "sure",
            "t",
            "true",
            "y",
            "yes",
        }:
            return True
        if val in {
            "0",
            "d",
            "disabled",
            "f",
            "false",
            "n",
            "no",
            "nope",
            "off",
            "r",
            "reject",
        }:
            return False
        if val in {"idc", "maybe", "random"}:
            return bool(random.randrange(2))  # nosec: B311
    if default is None:
        raise ValueError(f"Invalid bool value: {val!r}")
    return default


def str_to_set(string: str) -> set[str]:
    """Convert a string to a set of strings."""
    return {part.strip() for part in string.split(",") if part.strip()}


def strangle(string: str) -> float:
    """Convert a string to an angle."""
    hasher = sha1(string.encode("UTF-8"), usedforsecurity=False)
    return int.from_bytes(hasher.digest()[:2], "little") / (1 << 16) * 360


def time_function[  # noqa: D103
    # pylint: disable-next=invalid-name
    T, **P  # fmt: skip
](function: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> tuple[
    T, float
]:
    """Run the function and return the result and the time it took in seconds."""
    timer = Timer()
    return function(*args, **kwargs), timer.stop()


def time_to_str(spam: float) -> str:
    """Convert the time into a string with second precision."""
    int_time = int(spam)
    div_60 = int(int_time / 60)
    div_60_60 = int(div_60 / 60)

    return (
        f"{int(div_60_60 / 24)}d "
        f"{div_60_60 % 24}h "
        f"{div_60 % 60}min "
        f"{int_time % 60}s"
    )


@dataclass(order=True, frozen=True, slots=True)
class PageInfo:
    """The PageInfo class that is used for the subpages of a ModuleInfo."""

    name: str
    description: str
    path: None | str = None
    # keywords that can be used for searching
    keywords: tuple[str, ...] = field(default_factory=tuple)
    hidden: bool = False  # whether to hide this page info on the page
    short_name: None | str = None  # short name for the page


@dataclass(order=True, frozen=True, slots=True)
class ModuleInfo(PageInfo):
    """
    The ModuleInfo class adds handlers and subpages to the PageInfo.

    This gets created by every module to add the handlers.
    """

    handlers: tuple[Handler, ...] = field(default_factory=tuple[Handler, ...])
    sub_pages: tuple[PageInfo, ...] = field(default_factory=tuple)
    aliases: tuple[str, ...] | Mapping[str, str] = field(default_factory=tuple)
    required_background_tasks: Collection[BackgroundTask] = field(
        default_factory=frozenset
    )

    def get_keywords_as_str(self, path: str) -> str:
        """Get the keywords as comma-seperated string."""
        page_info = self.get_page_info(path)
        if self != page_info:
            return ", ".join((*self.keywords, *page_info.keywords))

        return ", ".join(self.keywords)

    def get_page_info(self, path: str) -> PageInfo:
        """Get the PageInfo of the specified path."""
        if self.path == path:
            return self

        for page_info in self.sub_pages:
            if page_info.path == path:
                return page_info

        return self



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the main module of an-website."""

from __future__ import annotations

import asyncio
import pathlib
import typing
from collections.abc import Mapping

import regex
from tornado.simple_httpclient import SimpleAsyncHTTPClient
from tornado.web import Application

from an_website import main, patches
from an_website.utils.base_request_handler import BaseRequestHandler
from an_website.utils.better_config_parser import BetterConfigParser

from . import (  # noqa: F401  # pylint: disable=unused-import
    PARENT_DIR,
    FetchCallable,
    app,
    assert_valid_redirect,
    fetch,
)


async def test_parsing_module_infos(
    fetch: FetchCallable,  # noqa: F811
    app: Application,  # noqa: F811
) -> None:
    """Tests about the module infos in main."""
    # pylint: disable=too-complex
    module_infos = app.settings["MODULE_INFOS"]
    # should get more than one module_info
    assert len(module_infos) > 0

    # module_infos should be sorted already; test that
    module_infos_list = list(module_infos)
    main.sort_module_infos(module_infos_list)
    assert module_infos == tuple(module_infos_list)

    arguments: set[tuple[str, str] | tuple[str, str, str]] = set()

    # tests about module infos
    for module_info in module_infos:
        if module_info.path is None:
            continue
        assert module_info.path.isascii()
        assert module_info.path.strip() == module_info.path
        assert not module_info.path.endswith("/") or module_info.path == "/"
        assert module_info.path in {module_info.path.lower(), "/LOLWUT"}

        for alias in module_info.aliases:
            assert alias.startswith("/")
            assert not alias.endswith("/")
            if module_info.path != "/chat":
                should_path: str = (
                    module_info.aliases[alias]
                    if isinstance(module_info.aliases, Mapping)
                    else module_info.path
                )
                kwargs: dict[str, typing.Any] = (
                    {}
                    if alias.isascii()
                    else {"httpclient": SimpleAsyncHTTPClient()}
                )
                await assert_valid_redirect(fetch, alias, should_path, **kwargs)

        if module_info.path != "/api/update":
            # check if at least one handler matches the path
            handler_matches_path = False
            for handler in module_info.handlers:
                if regex.fullmatch("(?i)" + handler[0], module_info.path):
                    handler_matches_path = True
                    break
            assert handler_matches_path

        if module_info.path in {
            "/chat",  # head not supported
            "/api/update",
            "/api/upload",
        }:
            continue
        if module_info.path == "/api/zufaelliger-text":
            arguments.update(
                (
                    (module_info.path, "*/*", "text/plain"),
                    (module_info.path, "text/plain"),
                    (module_info.path, "text/html"),
                )
            )
            continue
        if module_info.path.startswith("/api/"):
            should = (
                "text/plain"
                if module_info.path in {"/api/commitment", "/api/ping"}
                else "application/json"
            )
            arguments.update(
                (
                    (module_info.path, "*/*", should),
                    (module_info.path, should),
                )
            )
            continue
        arguments.update(
            (
                (module_info.path, "*/*", "text/html"),
                (module_info.path, "text/html"),
                (module_info.path, "text/markdown"),
                (module_info.path, "text/plain"),
                (module_info.path, "application/vnd.asozial.dynload+json"),
            )
        )

    # test all arguments
    await asyncio.gather(
        *(assert_valid_response(fetch, *argument) for argument in arguments)
    )

    # handlers should all be at least 3 long
    assert (
        min(
            len(handler)
            for handler in main.get_all_handlers(module_infos)
            if issubclass(handler[1], BaseRequestHandler)
        )
        == 3
    )


async def assert_valid_response(
    _fetch: FetchCallable, path: str, accept_header: str, content_type: str = ""
) -> None:
    """Assert that the response is valid."""
    follow_redirects = path in {
        "/redirect",
        "/zitat-des-tages",
    }
    valid_response_codes = {200, 503}
    if path == "/api/reports":
        valid_response_codes.add(404)
    elif path in {"/discord", "/wiki"}:
        valid_response_codes = {307}
    head_response = await _fetch(
        path,
        headers={"Accept": accept_header},
        method="HEAD",
        raise_error=False,
        follow_redirects=follow_redirects,
    )
    assert head_response.code in valid_response_codes
    assert not head_response.body
    get_response = await _fetch(
        path,
        headers={"Accept": accept_header},
        method="GET",
        raise_error=False,
        follow_redirects=follow_redirects,
    )
    assert get_response.code == head_response.code
    if get_response.code not in {204, 307, 308}:
        assert get_response.body.endswith(b"\n")
    for header in (
        "Content-Type",
        "Onion-Location",
        "Permissions-Policy",
    ):
        assert get_response.headers[header] == head_response.headers[header]
    assert get_response.headers["Content-Type"].startswith(
        content_type or accept_header
    )


def test_making_app() -> None:
    """Run the app making functions, to make sure they don't fail."""
    patches.apply()

    # read the example config, because it is always the same and should always work
    config = BetterConfigParser.from_path(
        pathlib.Path(PARENT_DIR, "config.ini.example")
    )

    application = main.make_app(config)

    main.apply_config_to_app(application, config)  # type: ignore[arg-type]

    # idk why; just to assert something lol
    assert application.settings["CONFIG"] == config  # type: ignore[union-attr]


if __name__ == "__main__":
    # test_parsing_module_infos()
    test_making_app()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the currency converter."""

from __future__ import annotations

import asyncio

from an_website.currency_converter import converter


def test_num_string_conversion() -> None:
    """Test the num_to_string and string_to_num conversion."""
    for num_str in ("0,50", "0,75", "1", "5,50", "10", "100"):
        _num: int = converter.string_to_num(num_str)
        assert converter.num_to_string(_num) == num_str

    assert converter.string_to_num("1,50") == 150
    assert converter.string_to_num("100xx") == 100 * 100
    assert converter.string_to_num("1,,50") == 150
    assert converter.string_to_num("100,,,,,,") == 100 * 100
    assert converter.string_to_num("1,50") == 150

    assert converter.num_to_string(150) == "1,50"
    assert converter.num_to_string(10000) == "100"

    for not_a_num in ("", ".", "..", "text"):
        assert not converter.string_to_num(not_a_num)


async def test_currency_conversion() -> None:
    """Test the currency conversion."""
    for _f in (0.5, 1, 2, 4, 8, 16, 32, 64, 128):
        val_dict = await converter.get_value_dict(int(_f * 100))
        for currency in ("euro", "mark", "ost", "schwarz"):
            assert val_dict[currency] == converter.string_to_num(
                val_dict[f"{currency}_str"]  # type: ignore[arg-type]
            )
            assert val_dict[f"{currency}_str"] == converter.num_to_string(
                val_dict[currency]  # type: ignore[arg-type]
            )
            assert str(val_dict[f"{currency}_str"]) in str(val_dict["text"])
    assert converter.convert(1) == (1, 2, 4, 20)
    assert converter.convert(2) == (2, 4, 8, 40)


if __name__ == "__main__":
    test_num_string_conversion()
    asyncio.run(test_currency_conversion())



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A 🆒 chat."""

from __future__ import annotations

from ..utils.utils import ModuleInfo
from .chat import (
    APIChatHandler,
    ChatWebSocketHandler,
    HTMLChatHandler,
    subscribe_to_redis_channel,
)


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (r"/emoji-chat", HTMLChatHandler),
            (r"/api/emoji-chat", APIChatHandler),
            (r"/websocket/emoji-chat", ChatWebSocketHandler),
        ),
        name="Emoji-Chat",
        description="Ein 🆒er Chat",
        path="/emoji-chat",
        keywords=("Emoji Chat",),
        required_background_tasks=(subscribe_to_redis_channel,),
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A Tornado template loader."""

from __future__ import annotations

import os.path
from importlib.resources.abc import Traversable
from typing import override

from tornado.template import BaseLoader, Template


class TemplateLoader(BaseLoader):
    """A Tornado template loader."""

    root: Traversable

    def __init__(self, root: Traversable, whitespace: None | str) -> None:
        """Initialize the template loader."""
        self.root = root
        super().__init__(whitespace=whitespace)

    @override
    def _create_template(self, name: str) -> Template:
        """Create a template from the given name."""
        return Template((self.root / name).read_bytes(), name=name, loader=self)

    @override
    def resolve_path(self, name: str, parent_path: str | None = None) -> str:
        """Resolve the template path.

        name is interpreted as relative to parent_path if it's not None.
        """
        if (
            # same check as in tornado.template.Loader
            parent_path
            and not parent_path.startswith("<")
            and not parent_path.startswith("/")
            and not name.startswith("/")
        ):
            return os.path.normpath(
                os.path.join(os.path.dirname(parent_path), name)
            )
        return name
